import json
import re
from pathlib import Path
from typing import Optional, List, Any
from pydantic import BaseModel, validator, Field, Extra
from cisco_sdwan.base.catalog import OpType, CATALOG_TAG_ALL, op_catalog_tags, op_catalog_commands, catalog_tags
from cisco_sdwan.base.models_base import filename_safe, DATA_DIR, ExtendedTemplate
from cisco_sdwan.tasks.utils import OpCmdOptions


# Model field validators
def validate_regex(regex: str) -> str:
    try:
        re.compile(regex)
    except (re.error, TypeError):
        raise ValueError(f'"{regex}" is not a valid regular expression.') from None
    return regex


def validate_workdir(workdir: str) -> str:
    if not Path(DATA_DIR, workdir).exists():
        raise ValueError(f'Work directory "{workdir}" not found.')

    return workdir


def validate_filename(filename: str) -> str:
    """ Validate file name. If filename is a path containing a directory, validate whether it exists.
    """
    file_path = Path(filename)
    if not file_path.parent.exists():
        raise ValueError(f'Directory for "{filename}" does not exist')

    # Also allow . on filename, on top of what's allowed by filename_safe
    if re.sub(r'\.', '_', file_path.name) != filename_safe(file_path.name):
        raise ValueError(
            f'Invalid name "{file_path.name}". Only alphanumeric characters, "-", "_", and "." are allowed.'
        )
    return filename


def validate_existing_file(filename: str) -> str:
    if not Path(filename).exists():
        raise ValueError(f'File "{filename}" not found.')

    return filename


def validate_ipv4(ipv4_str: str) -> str:
    if re.match(r'\d+(?:\.\d+){3}$', ipv4_str) is None:
        raise ValueError(f'"{ipv4_str}" is not a valid IPv4 address.')

    return ipv4_str


def validate_site_id(site_id: str) -> str:
    try:
        if not 0 <= int(site_id) <= 4294967295:
            raise ValueError()
    except ValueError:
        raise ValueError(f'"{site_id}" is not a valid site-id.') from None

    return site_id


def validate_op_cmd(op_type: OpType, cmd_list: List[str]) -> List[str]:
    full_command = ' '.join(cmd_list)
    pass_options = [
        len(cmd_list) == 1 and CATALOG_TAG_ALL in cmd_list,
        len(cmd_list) == 1 and set(cmd_list) <= op_catalog_tags(op_type),
        full_command in op_catalog_commands(op_type)
    ]
    if not any(pass_options):
        raise ValueError(f'"{full_command}" is not valid. '
                         f'Group options: {OpCmdOptions.tags(op_type)}. '
                         f'Command options: {OpCmdOptions.commands(op_type)}.')
    return cmd_list


def validate_json(json_str: str) -> str:
    try:
        json.loads(json_str)
    except json.JSONDecodeError as ex:
        raise ValueError(f'Invalid JSON data: {ex}') from None

    return json_str


catalog_tag_options = catalog_tags() | {CATALOG_TAG_ALL}


def validate_catalog_tag(tag: str) -> str:
    if tag not in catalog_tag_options:
        options = ', '.join(sorted(catalog_tag_options, key=lambda x: '' if x == CATALOG_TAG_ALL else x))
        raise ValueError(f'"{tag}" is not a valid tag. Available tags: {options}.')

    return tag


def validate_ext_template(template_str: str) -> str:
    # ExtendedTemplate will raise ValueError on validation failures
    ExtendedTemplate(template_str)('test')

    return template_str


def validate_version(version_str: str) -> str:
    # Development versions may follow this format: '20.1.999-98'
    if re.match(r'\d+([.-]\d+){1,3}$', version_str) is None:
        raise ValueError(f'"{version_str}" is not a valid version identifier.')

    return '.'.join(([str(int(v)) for v in version_str.replace('-', '.').split('.')] + ['0', ])[:2])


def const(default_value: Any) -> Field:
    """
    Defines a model field as constant. That is, it cannot be set to any value other than the default value
    """
    return Field(default_value, const=True)


# Models
class TaskArgs(BaseModel, extra=Extra.forbid):
    def __init__(self, **kwargs):
        # Dummy init used so PyCharm type checker can recognize TaskArgs parameters
        super().__init__(**kwargs)


class TableTaskArgs(TaskArgs):
    exclude: Optional[str] = None
    include: Optional[str] = None
    save_csv: Optional[str] = None
    save_json: Optional[str] = None

    # Validators
    _validate_regex = validator('exclude', 'include', allow_reuse=True)(validate_regex)
    _validate_filename = validator('save_csv', 'save_json', allow_reuse=True)(validate_filename)



import argparse
from typing import Union, Optional
from uuid import uuid4
from contextlib import suppress
from pydantic import validator
from cisco_sdwan.__version__ import __doc__ as title
from cisco_sdwan.base.rest_api import Rest
from cisco_sdwan.base.catalog import catalog_iter, CATALOG_TAG_ALL, ordered_tags
from cisco_sdwan.base.models_base import update_ids, ServerInfo, ExtendedTemplate
from cisco_sdwan.base.models_vmanage import DeviceTemplate, FeatureTemplate
from cisco_sdwan.base.processor import StopProcessorException, ProcessorException
from cisco_sdwan.migration import factory_cedge_aaa, factory_cedge_global
from cisco_sdwan.migration.feature_migration import FeatureProcessor
from cisco_sdwan.migration.device_migration import DeviceProcessor
from cisco_sdwan.tasks.utils import TaskOptions, existing_workdir_type, filename_type, version_type, ext_template_type
from cisco_sdwan.tasks.common import clean_dir, Task, TaskException
from cisco_sdwan.tasks.models import (TaskArgs, validate_workdir, validate_ext_template, validate_version,
                                      validate_filename)


@TaskOptions.register('migrate')
class TaskMigrate(Task):
    @staticmethod
    def parser(task_args, target_address=None):
        task_parser = argparse.ArgumentParser(description=f'{title}\nMigrate task:')
        task_parser.prog = f'{task_parser.prog} migrate'
        task_parser.formatter_class = argparse.RawDescriptionHelpFormatter

        task_parser.add_argument('scope', choices=['all', 'attached'],
                                 help='select whether to evaluate all feature templates, or only feature templates '
                                      'attached to device templates.')
        task_parser.add_argument('output', metavar='<directory>', type=filename_type,
                                 help='directory to save migrated templates')
        task_parser.add_argument('--no-rollover', action='store_true',
                                 help='by default, if the output directory already exists it is renamed using a '
                                      'rolling naming scheme. This option disables this automatic rollover.')
        task_parser.add_argument('--name', metavar='<name-regex>', type=ext_template_type, default='migrated_{name}',
                                 help='name-regex used to name the migrated templates (default: %(default)s). '
                                      'Variable {name} is replaced with the original template name. Sections of the '
                                      'original template name can be selected using the {name <regex>} syntax. Where '
                                      '<regex> is a regular expression that must contain at least one capturing group. '
                                      'Capturing groups identify sections of the original name to keep.')
        task_parser.add_argument('--from', metavar='<version>', type=version_type, dest='from_version', default='18.4',
                                 help='vManage version from source templates (default: %(default)s)')
        task_parser.add_argument('--to', metavar='<version>', type=version_type, dest='to_version', default='20.1',
                                 help='target vManage version for template migration (default: %(default)s)')
        task_parser.add_argument('--workdir', metavar='<directory>', type=existing_workdir_type,
                                 help='migrate will read from the specified directory instead of target vManage')

        return task_parser.parse_args(task_args)

    @staticmethod
    def is_api_required(parsed_args) -> bool:
        return parsed_args.workdir is None

    def runner(self, parsed_args, api: Optional[Rest] = None) -> Union[None, list]:
        source_info = f'Local workdir: "{parsed_args.workdir}"' if api is None else f'vManage URL: "{api.base_url}"'
        self.log_info('Migrate task: %s %s -> %s Local output dir: "%s"', source_info, parsed_args.from_version,
                      parsed_args.to_version, parsed_args.output)

        # Output directory must be empty for a new migration
        saved_output = clean_dir(parsed_args.output, max_saved=0 if parsed_args.no_rollover else 99)
        if saved_output:
            self.log_info('Previous migration under "%s" was saved as "%s"', parsed_args.output, saved_output)

        if api is None:
            backend = parsed_args.workdir
            local_info = ServerInfo.load(backend)
            server_version = local_info.server_version if local_info is not None else None
        else:
            backend = api
            server_version = backend.server_version

        try:
            # Load migration processors
            loaded_processors = {
                FeatureTemplate: FeatureProcessor.load(from_version=parsed_args.from_version,
                                                       to_version=parsed_args.to_version),
                DeviceTemplate: DeviceProcessor.load(from_version=parsed_args.from_version,
                                                     to_version=parsed_args.to_version)
            }
            self.log_info('Loaded template migration recipes')

            server_info = ServerInfo(server_version=parsed_args.to_version)
            if server_info.save(parsed_args.output):
                self.log_info('Saved vManage server information')

            id_mapping = {}  # {<old_id>: <new_id>}
            for tag in ordered_tags(CATALOG_TAG_ALL, reverse=True):
                self.log_info('Inspecting %s items', tag)

                for _, info, index_cls, item_cls in catalog_iter(tag, version=server_version):
                    item_index = self.index_get(index_cls, backend)
                    if item_index is None:
                        self.log_debug('Skipped %s, none found', info)
                        continue

                    name_set = {item_name for item_id, item_name in item_index}

                    is_bad_name = False
                    export_list = []
                    id_hint_map = {item_name: item_id for item_id, item_name in item_index}
                    for item_id, item_name in item_index:
                        item = self.item_get(item_cls, backend, item_id, item_name, item_index.need_extended_name)
                        if item is None:
                            self.log_error('Failed loading %s %s', info, item_name)
                            continue

                        with suppress(StopProcessorException):
                            item_processor = loaded_processors.get(item_cls)
                            if item_processor is None:
                                raise StopProcessorException()

                            self.log_debug('Evaluating %s %s', info, item_name)
                            if not item_processor.is_in_scope(item, migrate_all=(parsed_args.scope == 'all')):
                                self.log_debug('Skipping %s, migration not necessary', item_name)
                                raise StopProcessorException()

                            new_name = ExtendedTemplate(parsed_args.name)(item_name)
                            if not item_cls.is_name_valid(new_name):
                                self.log_error('New %s name is not valid: %s', info, new_name)
                                is_bad_name = True
                                raise StopProcessorException()
                            if new_name in name_set:
                                self.log_error('New %s name collision: %s -> %s', info, item_name, new_name)
                                is_bad_name = True
                                raise StopProcessorException()

                            name_set.add(new_name)

                            new_id = str(uuid4())
                            new_payload, trace_log = item_processor.eval(item, new_name, new_id)
                            for trace in trace_log:
                                self.log_debug('Processor: %s', trace)

                            if item.is_equal(new_payload):
                                self.log_debug('Skipping %s, no changes', item_name)
                                raise StopProcessorException()

                            new_item = item_cls(update_ids(id_mapping, new_payload))
                            id_mapping[item_id] = new_id
                            id_hint_map[new_name] = new_id

                            if item_processor.replace_original():
                                self.log_debug('Migrated replaces original: %s -> %s', item_name, new_name)
                                item = new_item
                            else:
                                self.log_debug('Migrated adds to original: %s + %s', item_name, new_name)
                                export_list.append(new_item)

                        export_list.append(item)

                    if is_bad_name:
                        raise TaskException(f'One or more new {info} names are not valid')

                    if not export_list:
                        self.log_info('No %s migrated', info)
                        continue

                    if issubclass(item_cls, FeatureTemplate):
                        for factory_default in (factory_cedge_aaa, factory_cedge_global):
                            if any(factory_default.name == elem.name for elem in export_list):
                                self.log_debug('Using existing factory %s %s', info, factory_default.name)
                                # Updating because device processor always use the built-in IDs
                                id_mapping[factory_default.uuid] = id_hint_map[factory_default.name]
                            else:
                                export_list.append(factory_default)
                                id_hint_map[factory_default.name] = factory_default.uuid
                                self.log_debug('Added factory %s %s', info, factory_default.name)

                    new_item_index = index_cls.create(export_list, id_hint_map)
                    if new_item_index.save(parsed_args.output):
                        self.log_info('Saved %s index', info)

                    for new_item in export_list:
                        if new_item.save(parsed_args.output, new_item_index.need_extended_name, new_item.name,
                                         id_hint_map[new_item.name]):
                            self.log_info('Saved %s %s', info, new_item.name)

        except (ProcessorException, TaskException) as ex:
            self.log_critical('Migration aborted: %s', ex)

        return


class MigrateArgs(TaskArgs):
    scope: str
    output: str
    no_rollover: bool = False
    name: str = 'migrated_{name}'
    from_version: str = '18.4'
    to_version: str = '20.1'
    workdir: Optional[str] = None

    # Validators
    _validate_filename = validator('output', allow_reuse=True)(validate_filename)
    _validate_name = validator('name', allow_reuse=True)(validate_ext_template)
    _validate_version = validator('from_version', 'to_version', allow_reuse=True)(validate_version)
    _validate_workdir = validator('workdir', allow_reuse=True)(validate_workdir)

    @validator('scope')
    def validate_scope(cls, v):
        scope_options = ('all', 'attached')
        if v not in scope_options:
            raise ValueError(f'"{v}" is not a valid scope. Options are: {", ".join(scope_options)}.')
        return v



from pathlib import Path
from cisco_sdwan.base.models_vmanage import FeatureTemplate
from cisco_sdwan.base.processor import ProcessorException

current_dir = Path(__file__).resolve().parent


def module_dir():
    return current_dir


# Load built-in factory default cedge_global and cedge_aaa
factory_cedge_global = FeatureTemplate.load(module_dir(), False, use_root_dir=False,
                                            item_name='Factory_Default_Global_CISCO_Template',
                                            item_id='300d7759-cc0a-4cd7-90c0-eb52adc27f2f')
if factory_cedge_global is None:
    raise ProcessorException('Unable to find factory default cEdge global template')

factory_cedge_aaa = FeatureTemplate.load(module_dir(), False, use_root_dir=False,
                                         item_name='Factory_Default_AAA_CISCO_Template',
                                         item_id='add276c5-45b0-4493-a559-5a07b15cbdeb')
if factory_cedge_aaa is None:
    raise ProcessorException('Unable to find factory default Cisco AAA template')



"""
 Sastre - Automation Tools for Cisco SD-WAN Powered by Viptela

 cisco_sdwan.base.rest_api
 This module implements the lower level REST API
"""
import requests
import urllib3
import json
from time import time
from typing import Optional


class Rest:
    def __init__(self, base_url: str, username: str, password: str, tenant_name: Optional[str] = None,
                 timeout: int = 20, verify: bool = False):
        self.base_url = base_url
        self.timeout = timeout
        self.verify = verify
        self.session = None
        self.server_facts = None
        self.is_tenant_scope = False

        if not verify:
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

        if not self.login(username, password, tenant_name):
            raise LoginFailedException(f'Login to {self.base_url} failed, check credentials')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.session is not None:
            # In 19.3, logging out actually de-authorize all sessions a user may have from the same IP address. For
            # instance browser windows open from the same laptop. This is fixed in 20.1.
            self.logout()
            self.session.close()

        return False

    def login(self, username: str, password: str, tenant_name: Optional[str]) -> bool:
        data = {
            'j_username': username,
            'j_password': password
        }

        session = requests.Session()
        response = session.post(f'{self.base_url}/j_security_check',
                                data=data, timeout=self.timeout, verify=self.verify)
        response.raise_for_status()

        if b'<html>' in response.content:
            return False

        self.session = session

        self.server_facts = self.get('client/server').get('data')
        if self.server_facts is None:
            raise RestAPIException('Could not retrieve vManage server information')

        # Token mechanism introduced in 19.2
        token = self.server_facts.get('CSRFToken')
        if token is not None:
            self.session.headers['X-XSRF-TOKEN'] = token

        self.session.headers['Content-Type'] = 'application/json'

        # Multi-tenant vManage with a provider account, insert vsessionid
        if tenant_name is not None:
            if not self.is_multi_tenant or not self.is_provider:
                raise BadTenantException('Tenant is only applicable to provider accounts in multi-tenant deployments')

            tenant_list = self.get('tenant').get('data')
            if tenant_list is None:
                raise RestAPIException('Could not retrieve vManage tenant list')

            for tenant in tenant_list:
                if tenant_name == tenant['name']:
                    session_id = self.post({}, 'tenant', tenant['tenantId'], 'vsessionid').get('VSessionId')
                    self.session.headers['VSessionId'] = session_id
                    self.is_tenant_scope = True
                    break
            else:
                raise BadTenantException(f'Invalid tenant: {tenant_name}')

        return True

    def logout(self):
        response = self.session.get(f'{self.base_url}/logout', params={'nocache': str(int(time()))})
        return response.status_code == requests.codes.ok

    @property
    def server_version(self):
        return self.server_facts.get('platformVersion')

    @property
    def is_multi_tenant(self):
        return self.server_facts.get('tenancyMode', '') == 'MultiTenant'

    @property
    def is_provider(self):
        return self.server_facts.get('userMode', '') == 'provider'

    def get(self, *path_entries, **params):
        response = self.session.get(self._url(*path_entries),
                                    params=params if params else None,
                                    timeout=self.timeout, verify=self.verify)
        raise_for_status(response)
        return response.json()

    def post(self, input_data, *path_entries):
        # With large input_data, vManage fails the post request if payload is encoded in compact form. Thus encoding
        # with indent=1.
        response = self.session.post(self._url(*path_entries), data=json.dumps(input_data, indent=1),
                                     timeout=self.timeout, verify=self.verify)
        raise_for_status(response)

        # POST may return an empty string, return None in this case
        return response.json() if response.text else None

    def put(self, input_data, *path_entries):
        # With large input_data, vManage fails the put request if payload is encoded in compact form. Thus encoding
        # with indent=1.
        response = self.session.put(self._url(*path_entries), data=json.dumps(input_data, indent=1),
                                    timeout=self.timeout, verify=self.verify)
        raise_for_status(response)

        # PUT may return an empty string, return None in this case
        return response.json() if response.text else None

    def delete(self, resource, key_value):
        response = self.session.delete(self._url(resource, key_value),
                                       timeout=self.timeout, verify=self.verify)
        return response.status_code == requests.codes.ok

    def _url(self, *path_entries):
        path = '/'.join(path.strip('/') for path in path_entries)

        return f'{self.base_url}/dataservice/{path}'


def raise_for_status(response):
    if response.status_code != requests.codes.ok:
        try:
            reply_data = response.json() if response.text else {}
        except json.decoder.JSONDecodeError:
            reply_data = {'error': {'message': 'Check user permissions'}} if response.status_code == 403 else {}

        details = reply_data.get("error", {}).get("details", "")
        raise RestAPIException(f'{response.reason} ({response.status_code}): '
                               f'{reply_data.get("error", {}).get("message", "Unspecified error message")}'
                               f'{": " if details else ""}{details} [{response.request.method} {response.url}]')


def is_version_newer(version_1, version_2):
    """
    Indicates whether one vManage version is newer than another. Compares only the first 2 digits from version
    because maintenance (i.e. 3rd digit) differences are not likely to create any incompatibility between REST API JSON
    payloads, thus not relevant in this context.

    Versions should be strings with fields separated by dots in format: <main>.<minor>.<maintenance>

    @param version_1: String containing first version
    @param version_2: String containing second version
    @return: True if version_2 is newer than version_1.
    """

    def parse(version_string):
        # Development versions may follow this format: '20.1.999-98'
        return ([int(v) for v in version_string.replace('-', '.').split('.')] + [0, ])[:2]

    return parse(version_2) > parse(version_1)


class RestAPIException(Exception):
    """ Exception for REST API errors """
    pass


class LoginFailedException(RestAPIException):
    """ Login failure """
    pass


class BadTenantException(RestAPIException):
    """ Provided tenant is invalid or not applicable """
    pass



"""
 Sastre - Automation Tools for Cisco SD-WAN Powered by Viptela

 cisco_sdwan.base.catalog
 This module implements vManage API Catalogs
"""
from typing import NamedTuple, Union, Optional, Iterator
from enum import Enum
from .models_base import IndexConfigItem, ConfigItem, RealtimeItem, BulkStateItem, BulkStatsItem
from .rest_api import is_version_newer

CATALOG_TAG_ALL = 'all'


# Catalog for configuration data items
class CatalogItem(NamedTuple):
    tag: str
    info: str
    index_cls: type
    item_cls: type
    min_version: Union[str, None]


_catalog = list()  # [(<tag>, <info>, <index_cls>, <item_cls>, <min_version>), ...]


# Catalog for operational data items
class OpType(Enum):
    STATS = BulkStatsItem
    STATE = BulkStateItem
    RT = RealtimeItem

    @classmethod
    def from_subclass(cls, op_cls):
        return cls(op_cls.__bases__[0])


class OpCatalogItem(NamedTuple):
    tag: str
    selector: str
    info: str
    op_cls: type
    min_version: Union[str, None]


_op_catalog = dict()  # {<OpType>: [(<tag>, <selector>, <info>, <op_cls>, <min_version>), ...]}

#
# Configuration catalog functions
#
# Order in which config items need to be deleted (i.e. reverse order in which they need to be pushed), considering
# their high-level dependencies.
_tag_dependency_list = [
    'template_device',
    'template_feature',
    'policy_vsmart',
    'policy_vedge',
    'policy_security',
    'policy_voice',
    'policy_customapp',
    'policy_definition',
    'policy_list',
]


def ordered_tags(tag: str, single: bool = False, reverse: bool = False) -> Iterator[str]:
    """
    Generator which yields the specified tag plus any 'child' tags (i.e. dependent tags), following the order in which
    items need to be removed based on their dependencies (e.g. template_device before template_feature). The overall
    order is defined by _tag_dependency_list.
    If special tag 'all' is used, all items from _tag_dependency_list are yielded.
    @param tag: tag string or 'all'
    @param single: Optional, when True only one (first) tag is yielded. Used mainly for convenience of the caller.
    @param reverse: If true, yield tags in reverse order
    @return: Selected tags in order, as per _tag_dependency_list
    """
    find_tag = (tag == CATALOG_TAG_ALL)
    for item in _tag_dependency_list if not reverse else reversed(_tag_dependency_list):
        if not find_tag:
            if item == tag:
                find_tag = True
            else:
                continue
        yield item

        if single:
            break


def register(tag: str, info: str, item_cls: type, min_version: Optional[str] = None):
    """
    Decorator used for registering config item index/handler classes with the catalog.
    The class being decorated needs to be a subclass of IndexConfigItem.
    @param tag: Tag string associated with this item. String 'all' is reserved and cannot be used.
    @param info: Item information used for logging purposes
    @param item_cls: The config item handler class, needs to be a subclass of ConfigItem
    @param min_version: (optional) Minimum vManage version that supports this catalog item.
    @return: decorator
    """

    def decorator(index_cls):
        if not isinstance(index_cls, type) or not issubclass(index_cls, IndexConfigItem):
            raise CatalogException(f'Invalid config item index class register attempt: {index_cls.__name__}')
        if not isinstance(item_cls, type) or not issubclass(item_cls, ConfigItem):
            raise CatalogException(
                f'Invalid config item class register attempt {index_cls.__name__}: {item_cls.__name__}'
            )
        if not isinstance(tag, str) or tag.lower() == CATALOG_TAG_ALL:
            raise CatalogException(f'Invalid tag provided for class {index_cls.__name__}: {tag}')
        if tag not in _tag_dependency_list:
            raise CatalogException(f'Unknown tag provided: {tag}')

        _catalog.append(CatalogItem(tag, info, index_cls, item_cls, min_version))

        return index_cls

    return decorator


def catalog_size() -> int:
    """
    Return number of entries in the catalog
    @return: integer
    """
    return len(_catalog)


def catalog_iter(*tags: str, version: Optional[str] = None) -> Iterator[tuple]:
    """
    Return an iterator of (<tag>, <info>, <index_cls>, <item_cls>) tuples matching the specified tag(s) and supported
    by vManage version.
    @param tags: tags indicating catalog entries to return
    @param version: Target vManage version. Only returns catalog items supported by the target vManage.
                    If not specified or None, version is not verified.
    @return: iterator of (<tag>, <info>, <index_cls>, <item_cls>) tuples from the catalog
    """

    def match_tags(clog_item):
        return CATALOG_TAG_ALL in tags or clog_item.tag in tags

    def match_version(clog_item):
        return clog_item.min_version is None or version is None or not is_version_newer(version, clog_item.min_version)

    return (
        (item.tag, item.info, item.index_cls, item.item_cls)
        for item in _catalog if match_tags(item) and match_version(item)
    )


def catalog_tags() -> set:
    """
    Return unique tags used by items registered with the catalog
    @return: Set of unique tags
    """
    return {entry.tag for entry in _catalog}


#
# Operational data catalog functions
#
def op_register(tag: str, selector: str, info: str, min_version: Optional[str] = None):
    """
    Decorator used for registering operational-data items with the op catalog.
    The class being decorated needs to be a subclass of OperationalItem.
    @param tag: Tag string associated with this item. String 'all' is reserved and cannot be used.
    @param selector: String used to further filter entries that match the tags.
    @param info: Item information used for logging purposes
    @param min_version: (optional) Minimum vManage version that supports this catalog item.
    @return: decorator
    """

    def decorator(op_cls):
        try:
            op_type = OpType.from_subclass(op_cls)
        except ValueError:
            raise CatalogException(f'Invalid operational-data class register attempt: {op_cls.__name__}') from None

        if not isinstance(tag, str) or tag.lower() == CATALOG_TAG_ALL:
            raise CatalogException(f'Invalid tag provided for class {op_cls.__name__}: {tag}')
        if not isinstance(selector, str) or selector.lower() == CATALOG_TAG_ALL:
            raise CatalogException(f'Invalid selector provided for class {op_cls.__name__}: {selector}')

        _op_catalog.setdefault(op_type, []).append(OpCatalogItem(tag, selector, info, op_cls, min_version))

        return op_cls

    return decorator


def op_catalog_size() -> int:
    """
    Return number of entries in the operational-data catalog
    @return: integer
    """
    return sum(len(entries) for entries in _op_catalog.values())


def op_catalog_iter(op_type: OpType, *tags: str, version: Optional[str] = None) -> Iterator[tuple]:
    """
    Return an iterator of (<info>, <op_cls>) tuples matching the specified tag(s), selector and supported
    by vManage version.
    @param op_type: OpType enum indicating type of operational-data
    @param tags: Tags to filter catalog entries to return. If 2 or more tags are provided, the last one is considered
                 a selector.
    @param version: Target vManage version. Only returns catalog items supported by the target vManage.
                    If not specified or None, version is not verified.
    @return: iterator of (<info>, <op_cls>) tuples from the operational-data catalog
    """
    if len(tags) > 1:
        group_list = tags[:-1]
        selector = tags[-1]
    else:
        group_list = tags
        selector = None

    def match_group(clog_item):
        return CATALOG_TAG_ALL in group_list or clog_item.tag in group_list

    def match_selector(clog_item):
        return selector is None or clog_item.selector == selector

    def match_version(clog_item):
        return clog_item.min_version is None or version is None or not is_version_newer(version, clog_item.min_version)

    return (
        (item.info, item.op_cls)
        for item in _op_catalog.get(op_type, []) if match_group(item) and match_selector(item) and match_version(item)
    )


def op_catalog_tags(op_type: OpType) -> set:
    """
    Return unique tags used by items registered with the operational-data catalog group
    @param op_type: OpType enum indicating type of operational-data
    @return: Set of unique tags
    """
    return {entry.tag for entry in _op_catalog.get(op_type, [])}


def op_catalog_commands(op_type: OpType) -> set:
    """
    Return set of commands registered with the operational-data catalog group. These are the combination of tags and
    selectors
    @param op_type: OpType enum indicating type of operational-data
    @return: Set of commands
    """
    return {f'{entry.tag} {entry.selector}' for entry in _op_catalog.get(op_type, [])}


class CatalogException(Exception):
    """ Exception for config item catalog errors """
    pass



import argparse
from typing import Union, Optional, List
from pydantic import validator
from cisco_sdwan.__version__ import __doc__ as title
from cisco_sdwan.base.rest_api import Rest, RestAPIException
from cisco_sdwan.base.catalog import catalog_iter, CATALOG_TAG_ALL
from cisco_sdwan.base.models_base import ServerInfo
from cisco_sdwan.base.models_vmanage import (DeviceConfig, DeviceConfigRFS, DeviceTemplate, DeviceTemplateAttached,
                                             DeviceTemplateValues, EdgeInventory, ControlInventory, EdgeCertificate)
from cisco_sdwan.tasks.utils import TaskOptions, TagOptions, filename_type, regex_type, default_workdir
from cisco_sdwan.tasks.common import regex_search, clean_dir, Task
from cisco_sdwan.tasks.models import TaskArgs, validate_regex, validate_filename, validate_catalog_tag


@TaskOptions.register('backup')
class TaskBackup(Task):
    @staticmethod
    def parser(task_args, target_address=None):
        task_parser = argparse.ArgumentParser(description=f'{title}\nBackup task:')
        task_parser.prog = f'{task_parser.prog} backup'
        task_parser.formatter_class = argparse.RawDescriptionHelpFormatter

        task_parser.add_argument('--workdir', metavar='<directory>', type=filename_type,
                                 default=default_workdir(target_address),
                                 help='backup destination (default: %(default)s)')
        task_parser.add_argument('--no-rollover', action='store_true',
                                 help='by default, if workdir already exists (before a new backup is saved) the old '
                                      'workdir is renamed using a rolling naming scheme. This option disables this '
                                      'automatic rollover.')
        task_parser.add_argument('--save-running', action='store_true',
                                 help='include the running config from each node to the backup. This is useful for '
                                      'reference or documentation purposes. It is not needed by the restore task.')
        mutex = task_parser.add_mutually_exclusive_group()
        mutex.add_argument('--regex', metavar='<regex>', type=regex_type,
                           help='regular expression matching item names to backup, within selected tags.')
        mutex.add_argument('--not-regex', metavar='<regex>', type=regex_type,
                           help='regular expression matching item names NOT to backup, within selected tags.')
        task_parser.add_argument('tags', metavar='<tag>', nargs='+', type=TagOptions.tag,
                                 help='one or more tags for selecting items to be backed up. Multiple tags should be '
                                      f'separated by space. Available tags: {TagOptions.options()}. Special tag '
                                      f'"{CATALOG_TAG_ALL}" selects all items, including WAN edge certificates and '
                                      'device configurations.')
        return task_parser.parse_args(task_args)

    def runner(self, parsed_args, api: Optional[Rest] = None) -> Union[None, list]:
        self.log_info('Backup task: vManage URL: "%s" -> Local workdir: "%s"', api.base_url, parsed_args.workdir)

        # Backup workdir must be empty for a new backup
        saved_workdir = clean_dir(parsed_args.workdir, max_saved=0 if parsed_args.no_rollover else 99)
        if saved_workdir:
            self.log_info('Previous backup under "%s" was saved as "%s"', parsed_args.workdir, saved_workdir)

        target_info = ServerInfo(server_version=api.server_version)
        if target_info.save(parsed_args.workdir):
            self.log_info('Saved vManage server information')

        if parsed_args.save_running:
            self.save_running_configs(api, parsed_args.workdir)

        # Backup items not registered to the catalog, but to be included when tag is 'all'
        if CATALOG_TAG_ALL in parsed_args.tags:
            edge_certs = EdgeCertificate.get(api)
            if edge_certs is None:
                self.log_error('Failed backup WAN edge certificates')
            elif edge_certs.save(parsed_args.workdir):
                self.log_info('Saved WAN edge certificates')

        # Backup items registered to the catalog
        for _, info, index_cls, item_cls in catalog_iter(*parsed_args.tags, version=api.server_version):
            item_index = index_cls.get(api)
            if item_index is None:
                self.log_debug('Skipped %s, item not supported by this vManage', info)
                continue
            if item_index.save(parsed_args.workdir):
                self.log_info('Saved %s index', info)

            regex = parsed_args.regex or parsed_args.not_regex
            matched_item_iter = (
                (item_id, item_name) for item_id, item_name in item_index
                if regex is None or regex_search(regex, item_name, inverse=parsed_args.regex is None)
            )
            for item_id, item_name in matched_item_iter:
                item = item_cls.get(api, item_id)
                if item is None:
                    self.log_error('Failed backup %s %s', info, item_name)
                    continue
                if item.save(parsed_args.workdir, item_index.need_extended_name, item_name, item_id):
                    self.log_info('Done %s %s', info, item_name)

                # Special case for DeviceTemplateAttached and DeviceTemplateValues
                if isinstance(item, DeviceTemplate):
                    devices_attached = DeviceTemplateAttached.get(api, item_id)
                    if devices_attached is None:
                        self.log_error('Failed backup %s %s attached devices', info, item_name)
                        continue
                    if devices_attached.save(parsed_args.workdir, item_index.need_extended_name, item_name, item_id):
                        self.log_info('Done %s %s attached devices', info, item_name)
                    else:
                        self.log_debug('Skipped %s %s attached devices, none found', info, item_name)
                        continue

                    try:
                        uuid_list = [uuid for uuid, _ in devices_attached]
                        values = DeviceTemplateValues(api.post(DeviceTemplateValues.api_params(item_id, uuid_list),
                                                               DeviceTemplateValues.api_path.post))
                        if values.save(parsed_args.workdir, item_index.need_extended_name, item_name, item_id):
                            self.log_info('Done %s %s values', info, item_name)
                    except RestAPIException as ex:
                        self.log_error('Failed backup %s %s values: %s', info, item_name, ex)

        return

    def save_running_configs(self, api: Optional[Rest], workdir: str) -> None:
        inventory_list = [(ControlInventory.get(api), 'controller')]
        if not api.is_provider or api.is_tenant_scope:
            inventory_list.append((EdgeInventory.get(api), 'WAN edge'))

        for inventory, info in inventory_list:
            if inventory is None:
                self.log_error('Failed retrieving %s inventory', info)
                continue

            for uuid, _, hostname, _ in inventory.extended_iter():
                if hostname is None:
                    self.log_debug('Skipping %s, no hostname', uuid)
                    continue

                for item, config_type in ((DeviceConfig.get(api, DeviceConfig.api_params(uuid)), 'CFS'),
                                          (DeviceConfigRFS.get(api, DeviceConfigRFS.api_params(uuid)), 'RFS')):
                    if item is None:
                        self.log_error('Failed backup %s device configuration %s', config_type, hostname)
                        continue
                    if item.save(workdir, item_name=hostname, item_id=uuid):
                        self.log_info('Done %s device configuration %s', config_type, hostname)


class BackupArgs(TaskArgs):
    regex: Optional[str] = None
    not_regex: Optional[str] = None
    no_rollover: bool = False
    save_running: bool = False
    workdir: str
    tags: List[str]

    # Validators
    _validate_regex = validator('regex', 'not_regex', allow_reuse=True)(validate_regex)
    _validate_workdir = validator('workdir', allow_reuse=True)(validate_filename)
    _validate_tags = validator('tags', each_item=True, allow_reuse=True)(validate_catalog_tag)



"""
 Sastre - Automation Tools for Cisco SD-WAN Powered by Viptela

 cisco_sdwan.tasks.common
 This module implements supporting classes and functions for tasks
"""
import logging
import time
import csv
import re
import json
from pathlib import Path
from shutil import rmtree
from collections import namedtuple
from typing import List, Tuple, Iterator, Union, Optional, Any, Iterable, Type, TypeVar, Sequence
from cisco_sdwan.base.rest_api import Rest, RestAPIException
from cisco_sdwan.base.models_base import DATA_DIR
from cisco_sdwan.base.models_vmanage import (DeviceTemplate, DeviceTemplateValues, DeviceTemplateAttached,
                                             DeviceTemplateAttach, DeviceTemplateCLIAttach, DeviceModeCli,
                                             ActionStatus, PolicyVsmartStatus, PolicyVsmartStatusException,
                                             PolicyVsmartActivate, PolicyVsmartIndex, PolicyVsmartDeactivate,
                                             Device)

T = TypeVar('T')


def regex_search(regex: str, *fields: str, inverse: bool = False) -> bool:
    """
    Execute regular expression search on provided fields. Match fields in the order provided. Behavior is determined
    by the inverse field. With inverse False (default), returns True (i.e. match) if pattern matches any field. When
    inverse is True, returns True if pattern does not match all fields
    @param regex: Pattern to match
    @param fields: One or more strings to match
    @param inverse: False (default), or True to invert the match behavior.
    @return: True if a match is found on any field, False otherwise.
    """
    op_fn = all if inverse else any  # Logical AND across all fields, else logical OR
    return op_fn(inverse ^ bool(re.search(regex, match_field)) for match_field in fields)


class Tally:
    def __init__(self, *counters):
        self._tally = {counter: 0 for counter in counters}

    def __getattr__(self, counter):
        return self._tally[counter]

    def incr(self, counter):
        self._tally[counter] += 1


class TableFilter:
    def __init__(self, regex: str, column: Optional[int] = None, inverse: bool = False):
        """
        Table row filter used by 'Table.filtered'
        @param regex: Pattern to match
        @param column: Column that the filter should match. When None, filter matches if any column matches the pattern.
        @param inverse: If False (default) the filter returns True on matches. That is, a row that matches is allowed
                        by the filter. When True, this is inverted; row that does not match is allowed by the filter.
        """
        if column is not None and column < 0:
            raise ValueError('Column value, when provided, must be a positive integer')

        self.regex_pattern = re.compile(regex)
        self.column = column
        self.inverse = inverse

    def __call__(self, table_row: tuple) -> bool:
        """
        Callable used by 'Table.filtered' to evaluate whether a row should be allowed.
        @return: True if row is allowed by the filter. False otherwise.
        """
        def match_column(column_value) -> bool:
            if column_value is None:
                return True

            return self.regex_pattern.search(str(column_value)) is not None

        # Table row is None when that is a marker row
        if table_row is None:
            return True

        # When column is not provided, any column match is a row match
        if self.column is None:
            return self.inverse ^ any(match_column(cell_value) for cell_value in table_row)

        # With column provided, match on that particular column is a row match
        return self.inverse ^ match_column(table_row[self.column])


class Table:
    DECIMAL_DIGITS = 1  # Number of decimal digits for float values

    def __init__(self, *columns: str, name: Optional[str] = None, meta: Optional[str] = None) -> None:
        self.header = tuple(columns)
        self.name = name
        self.meta = meta
        self._row_class = namedtuple('Row', (f'column_{i}' for i in range(len(columns))))
        self._rows = list()

    def filtered(self, *filter_fns: TableFilter):
        """
        Returns a new Table instance constructed off this Table object, with its rows filtered by evaluation of
        one or more TableFilters. A row is included if it is allowed by all TableFilters.
        @param filter_fns: one or more TableFilter instances
        @return: New Table instance containing the filtered rows
        """
        new_table = Table(*self.header, name=self.name, meta=self.meta)
        new_table._rows = [row for row in self if all(filter_fn(row) for filter_fn in filter_fns)]

        return new_table

    @staticmethod
    def process_value(value):
        return round(value, Table.DECIMAL_DIGITS) if isinstance(value, float) else value

    def add(self, *row_values):
        self._rows.append(self._row_class(*map(self.process_value, row_values)))

    def add_marker(self):
        self._rows.append(None)

    def extend(self, row_values_iter):
        self._rows.extend(self._row_class(*map(self.process_value, row_values)) for row_values in row_values_iter)

    def __iter__(self):
        return iter(self._rows)

    def __len__(self):
        total_len = len(self._rows) - self._rows.count(None)
        return total_len if total_len > 0 else 0

    def __str__(self):
        return '\n'.join(self.pretty_iter())

    def _column_max_width(self, index):
        def cell_length(cell_value):
            return len(str(cell_value))

        return max(
            cell_length(self.header[index]),
            max((cell_length(row[index]) for row in self._rows if row is not None)) if len(self) > 0 else 0
        )

    def pretty_iter(self):
        def cell_format(width, value):
            return f' {str(value):{width - 2}} '

        def border(line_ch: str, int_edge_ch: str = '+', ext_edge_ch: str = '+') -> str:
            return ext_edge_ch + int_edge_ch.join((line_ch * col_width for col_width in col_width_list)) + ext_edge_ch

        col_width_list = [2 + self._column_max_width(index) for index in range(len(self.header))]
        border_line = border('-')
        header_border_line = border('=', '=')

        if self.name is not None:
            yield f"*** {self.name} ***"

        yield header_border_line
        yield '|' + '|'.join(cell_format(width, value) for width, value in zip(col_width_list, self.header)) + '|'
        yield header_border_line

        done_content_row = False
        for row in self._rows:
            if row is not None:
                done_content_row = True
                yield '|' + '|'.join(cell_format(width, value) for width, value in zip(col_width_list, row)) + '|'
            elif done_content_row:
                done_content_row = False
                yield border_line

        if done_content_row:
            yield border_line

    def dict(self) -> dict:
        table_dict = {
            "header": {
                "name": self.name or "",
                "title": {column_id: title for column_id, title in zip(self._row_class._fields, self.header)}
            },
            "data": [row._asdict() for row in self._rows if row is not None]
        }
        return table_dict

    def json(self, **dumps_kwargs: Any) -> str:
        return json.dumps(self.dict(), **dumps_kwargs)

    def save(self, filename):
        with open(filename, 'w', newline='') as csv_file:
            writer = csv.writer(csv_file)
            writer.writerow(self.header)
            writer.writerows(row for row in self._rows if row is not None)


def get_table_filters(exclude_regex: Optional[str], include_regex: Optional[str]) -> Sequence[TableFilter]:
    filters = []
    if exclude_regex is not None:
        filters.append(TableFilter(exclude_regex, inverse=True))
    if include_regex is not None:
        filters.append(TableFilter(include_regex))

    return filters


def filtered_tables(tables: Sequence[Table], *filter_fns: TableFilter) -> Sequence[Table]:
    if not filter_fns:
        return tables

    filtered_table_iter = (table.filtered(*filter_fns) for table in tables)

    return [filtered_table for filtered_table in filtered_table_iter if filtered_table]


class DryRunReport:
    def __init__(self):
        self._entries: List[str] = []

    def add(self, entry: str) -> None:
        self._entries.append(entry)

    def __iter__(self) -> Iterator[str]:
        return iter(self._entries)

    def render(self) -> Iterator[str]:
        yield ""
        yield "### Dry-run actions preview ###"
        yield ""
        yield from (f"    {entry}" for entry in self)
        yield ""

    def __str__(self) -> str:
        return '\n'.join(self.render())


class Task:
    # Configuration parameters for wait_actions
    ACTION_INTERVAL = 10  # seconds
    ACTION_TIMEOUT = 1800  # 30 minutes

    SAVINGS_FACTOR = 1
    DRYRUN_LEVELS = {'info', 'warning', 'error', 'critical'}

    def __init__(self):
        self.log_count = Tally('debug', 'info', 'warning', 'error', 'critical')
        self.is_dryrun = False
        self.dryrun_report = DryRunReport()

    def log_debug(self, msg: str, *args, dryrun: bool = True) -> None:
        self._log('debug', msg, *args, dryrun=dryrun)

    def log_info(self, msg: str, *args, dryrun: bool = True) -> None:
        self._log('info', msg, *args, dryrun=dryrun)

    def log_warning(self, msg: str, *args, dryrun: bool = True) -> None:
        self._log('warning', msg, *args, dryrun=dryrun)

    def log_error(self, msg: str, *args, dryrun: bool = True) -> None:
        self._log('error', msg, *args, dryrun=dryrun)

    def log_critical(self, msg: str, *args, dryrun: bool = True) -> None:
        self._log('critical', msg, *args, dryrun=dryrun)

    def _log(self, level: str, msg: str, *args, dryrun: bool) -> None:
        """
        Logs a message
        @param level: Logging level
        @param msg: Log message
        @param args: Optional args to replace in msg via % operator
        @param dryrun: Whether to include this message to the dryrun report. Messages are added to the dryrun report if
                       this flag is True, the task is in dryrun mode and the level is in DRYRUN_LEVELS.
        """
        getattr(logging.getLogger(type(self).__name__), level)(f"DRY-RUN: {msg}" if self.is_dryrun else msg, *args)
        self.log_count.incr(level)

        if self.is_dryrun and dryrun and level in Task.DRYRUN_LEVELS:
            self.dryrun_report.add(msg % args if args else msg)

    def outcome(self, success_msg, failure_msg):
        msg_list = list()
        if self.log_count.critical:
            msg_list.append(f'{self.log_count.critical} critical')
        if self.log_count.error:
            msg_list.append(f'{self.log_count.error} errors')
        if self.log_count.warning:
            msg_list.append(f'{self.log_count.warning} warnings')

        msg = failure_msg if len(msg_list) > 0 else success_msg
        return msg.format(tally=', '.join(msg_list))

    @property
    def savings(self):
        """
        Estimate number of hours saved when running this task, when compared with performing the same steps manually.
        """
        return self.SAVINGS_FACTOR * self.log_count.info / 60

    @staticmethod
    def parser(task_args, **kwargs):
        raise NotImplementedError()

    @staticmethod
    def is_api_required(parsed_args) -> bool:
        return True

    def runner(self, parsed_args, api: Optional[Rest] = None) -> Union[None, list]:
        """
        Execute the task. If the task has some output to the user, that is returned via a list of objects. Objects in
        that list need to implement the __str__ method. If the task generates no output to the user None is returned.
        """
        raise NotImplementedError()

    def index_iter(self, backend, catalog_entry_iter):
        """
        Return an iterator of indexes loaded from backend. If backend is a Rest API instance, indexes are loaded
        from remote vManage via API. Otherwise, items are loaded from local backup under the backend directory.
        @param backend: Rest api instance or directory name
        @param catalog_entry_iter: An iterator of CatalogEntry
        @return: Iterator of (<tag>, <info>, <index>, <item_cls>)
        """
        is_api = isinstance(backend, Rest)

        def load_index(index_cls, info):
            index = index_cls.get(backend) if is_api else index_cls.load(backend)
            self.log_debug(f'{"No" if index is None else "Loaded"} {"remote" if is_api else "local"} {info} index')
            return index

        all_index_iter = (
            (tag, info, load_index(index_cls, info), item_cls)
            for tag, info, index_cls, item_cls in catalog_entry_iter
        )
        return ((tag, info, index, item_cls) for tag, info, index, item_cls in all_index_iter if index is not None)

    @staticmethod
    def item_get(item_cls: Type[T], backend: Union[Rest, str],
                 item_id: str, item_name: str, ext_name: bool) -> Union[T, None]:
        if isinstance(backend, Rest):
            return item_cls.get(backend, item_id)
        else:
            return item_cls.load(backend, ext_name, item_name, item_id)

    @staticmethod
    def index_get(index_cls: Type[T], backend: Union[Rest, str]) -> Union[T, None]:
        return index_cls.get(backend) if isinstance(backend, Rest) else index_cls.load(backend)

    def attach_template_data(self, api: Rest, workdir: str, ext_name: bool, templates_iter: Iterator[tuple],
                             target_uuid_set: Optional[set] = None) -> Tuple[list, bool]:
        """
        Prepare data for template attach considering local backup as the source of truth (i.e. where input values are)
        @param api: Instance of Rest API
        @param workdir: Directory containing saved items
        @param ext_name: Boolean passed to .load methods indicating whether extended item names should be used.
        @param templates_iter: Iterator of (<template_name>, <saved_template_id>, <target_template_id>)
        @param target_uuid_set: (optional) Set of existing device uuids on target node.
                                When provided, attach only devices that were previously attached (on saved) and are on
                                target node but are not yet attached.
                                When absent, re-attach all currently attached devices on target.
        @return: Tuple containing attach data (<template input list>, <isEdited>)
        """

        def load_template_input(template_name: str, saved_id: str, target_id: str) -> Union[list, None]:
            if target_id is None:
                self.log_debug(f'Skip {template_name}, saved template is not on target node')
                return None

            saved_values = DeviceTemplateValues.load(workdir, ext_name, template_name, saved_id)
            if saved_values is None:
                self.log_error(f'DeviceTemplateValues file not found: {template_name}, {saved_id}')
                return None
            if saved_values.is_empty:
                self.log_debug(f'Skip {template_name}, saved template has no attachments')
                return None

            target_attached_uuid_set = {uuid for uuid, _ in DeviceTemplateAttached.get_raise(api, target_id)}
            if target_uuid_set is None:
                allowed_uuid_set = target_attached_uuid_set
            else:
                saved_attached = DeviceTemplateAttached.load(workdir, ext_name, template_name, saved_id)
                if saved_attached is None:
                    self.log_error(f'DeviceTemplateAttached file not found: {template_name}, {saved_id}')
                    return None
                saved_attached_uuid_set = {uuid for uuid, _ in saved_attached}
                allowed_uuid_set = target_uuid_set & saved_attached_uuid_set - target_attached_uuid_set

            input_list = saved_values.input_list(allowed_uuid_set)
            if len(input_list) == 0:
                self.log_debug(f'Skip {template_name}, no devices to attach')
                return None

            return input_list

        def is_template_cli(template_name: str, saved_id: str) -> bool:
            return DeviceTemplate.load(workdir, ext_name, template_name, saved_id, raise_not_found=True).is_type_cli

        template_input_list = [
            (name, target_id, load_template_input(name, saved_id, target_id), is_template_cli(name, saved_id))
            for name, saved_id, target_id in templates_iter
        ]
        return template_input_list, target_uuid_set is None

    @staticmethod
    def reattach_template_data(api: Rest, templates_iter: Iterator[tuple]) -> Tuple[list, bool]:
        """
        Prepare data for template reattach considering vManage as the source of truth (i.e. where input values are)
        @param api: Instance of Rest API
        @param templates_iter: Iterator of (<template_name>, <target_template_id>)
        @return: Tuple containing attach data (<template input list>, <isEdited>)
        """

        def get_template_input(template_id):
            uuid_list = [uuid for uuid, _ in DeviceTemplateAttached.get_raise(api, template_id)]
            values = DeviceTemplateValues(api.post(DeviceTemplateValues.api_params(template_id, uuid_list),
                                                   DeviceTemplateValues.api_path.post))
            return values.input_list()

        def is_template_cli(template_id):
            return DeviceTemplate.get_raise(api, template_id).is_type_cli

        template_input_list = [
            (template_name, template_id, get_template_input(template_id), is_template_cli(template_id))
            for template_name, template_id in templates_iter
        ]
        return template_input_list, True

    def attach(self, api: Rest, template_input_list: List[tuple], is_edited: bool, *, chunk_size: int = 200,
               log_context: str, raise_on_failure: bool = True) -> int:
        """
        Attach device templates to devices
        @param api: Instance of Rest API
        @param template_input_list: List containing payload for template attachment
        @param is_edited: Boolean corresponding to the isEdited tag in the template attach payload
        @param chunk_size: Maximum number of device attachments per request
        @param raise_on_failure: If True, raise exception on action failures
        @param log_context: Message to log during wait actions
        @return: Number of attachment requests processed
        """

        def grouper(attach_cls, request_list):
            while True:
                section_dict = yield from chopper(chunk_size)
                if not section_dict:
                    continue

                request_details = (
                    f"{template_name} ({', '.join(DeviceTemplateValues.input_list_devices(input_list))})"
                    for template_name, key_dict in section_dict.items() for _, input_list in key_dict.items()
                )
                self.log_info(f'Template attach: {", ".join(request_details)}')

                if not self.is_dryrun:
                    template_input_iter = (
                        (template_id, input_list)
                        for key_dict in section_dict.values() for template_id, input_list in key_dict.items()
                    )
                    action_worker = attach_cls(
                        api.post(attach_cls.api_params(template_input_iter, is_edited), attach_cls.api_path.post)
                    )
                    self.log_debug(f'Device template attach requested: {action_worker.uuid}')
                    self.wait_actions(api, [(action_worker, ', '.join(section_dict))], log_context, raise_on_failure)

                request_list.append(...)

        def feeder(attach_cls, attach_data_iter):
            attach_reqs = []
            group = grouper(attach_cls, attach_reqs)
            next(group)
            for template_name, template_id, input_list in attach_data_iter:
                for input_entry in input_list:
                    group.send((template_name, template_id, input_entry))
            group.send(None)

            return attach_reqs

        # Attach requests for feature-based device templates
        feature_based_iter = ((template_name, template_id, input_list)
                              for template_name, template_id, input_list, is_cli in template_input_list
                              if input_list is not None and not is_cli)
        feature_based_reqs = feeder(DeviceTemplateAttach, feature_based_iter)

        # Attach Requests for cli device templates
        cli_based_iter = ((template_name, template_id, input_list)
                          for template_name, template_id, input_list, is_cli in template_input_list
                          if input_list is not None and is_cli)
        cli_based_reqs = feeder(DeviceTemplateCLIAttach, cli_based_iter)

        return len(feature_based_reqs + cli_based_reqs)

    def detach(self, api: Rest, template_iter: Iterator[tuple], device_map: Optional[dict] = None, *,
               chunk_size: int = 200, log_context: str, raise_on_failure: bool = True) -> int:
        """
        Detach devices from device templates
        @param api: Instance of Rest API
        @param template_iter: An iterator of (<template id>, <template name>) tuples containing templates to detach
        @param device_map: {<uuid>: <name>, ...} dict containing allowed devices for detach. If None, all attached
                           devices are detached.
        @param chunk_size: Maximum number of device detachments per request
        @param raise_on_failure: If True, raise exception on action failures
        @param log_context: Message to log during wait actions
        @return: Number of detach requests processed
        """

        def grouper(request_list):
            while True:
                section_dict = yield from chopper(chunk_size)
                if not section_dict:
                    continue

                wait_list = []
                for device_type, key_dict in section_dict.items():
                    request_details = (
                        f"{t_name} ({', '.join(device_map.get(device_id, '-') for device_id in device_id_list)})"
                        for t_name, device_id_list in key_dict.items()
                    )
                    self.log_info(f'Template detach: {", ".join(request_details)}')

                    if not self.is_dryrun:
                        id_list = (device_id for device_id_list in key_dict.values() for device_id in device_id_list)
                        action_worker = DeviceModeCli(
                            api.post(DeviceModeCli.api_params(device_type, *id_list), DeviceModeCli.api_path.post)
                        )
                        wait_list.append((action_worker, ', '.join(key_dict)))
                        self.log_debug(f'Device template attach requested: {action_worker.uuid}')

                    request_list.append(...)

                if wait_list:
                    self.wait_actions(api, wait_list, log_context, raise_on_failure)

        detach_reqs = []
        group = grouper(detach_reqs)
        next(group)

        if device_map is None:
            device_map = dict(device_iter(api))

        for template_id, template_name in template_iter:
            devices_attached = DeviceTemplateAttached.get(api, template_id)
            if devices_attached is None:
                self.log_warning(f'Failed to retrieve {template_name} attached devices from vManage')
                continue
            for uuid, personality in devices_attached:
                if uuid in device_map:
                    group.send((personality, template_name, uuid))
        group.send(None)

        return len(detach_reqs)

    def activate_policy(self, api: Rest, policy_id: Optional[str], policy_name: Optional[str],
                        is_edited: bool = False) -> List[tuple]:
        """
        @param api: Instance of Rest API
        @param policy_id: ID of policy to activate
        @param policy_name: Name of policy to activate
        @param is_edited: (optional) When true it indicates reactivation of an already active policy (e.x. due to
                                     in-place modifications)
        @return: List of worker actions to monitor [(<action_worker>, <template_name>), ...]
        """
        action_list = []
        if policy_id is None or policy_name is None:
            # No policy is active or policy not on target vManage
            return action_list

        try:
            PolicyVsmartStatus.get_raise(api).raise_for_status()
        except (RestAPIException, PolicyVsmartStatusException):
            self.log_debug('vSmarts not in vManage mode or otherwise not ready to have policy activated')
        else:
            action_worker = PolicyVsmartActivate(
                api.post(PolicyVsmartActivate.api_params(is_edited), PolicyVsmartActivate.api_path.post, policy_id)
            )
            self.log_debug(f'Policy activate requested: {action_worker.uuid}')
            action_list.append((action_worker, policy_name))

        return action_list

    def deactivate_policy(self, api: Rest) -> List[tuple]:
        action_list = []
        item_id, item_name = PolicyVsmartIndex.get_raise(api).active_policy
        if item_id is not None and item_name is not None:
            action_worker = PolicyVsmartDeactivate(api.post({}, PolicyVsmartDeactivate.api_path.post, item_id))
            self.log_debug(f'Policy deactivate requested: {action_worker.uuid}')
            action_list.append((action_worker, item_name))

        return action_list

    def wait_actions(self, api: Rest, action_list: List[tuple], log_context: str, raise_on_failure: bool) -> bool:
        """
        Wait for actions in action_list to complete
        @param api: Instance of Rest API
        @param action_list: [(<action_worker>, <action_info>), ...]. Where <action_worker> is an instance of ApiItem and
                            <action_info> is a str with information about the action. Action_info can be None, in which
                            case no messages are logged for individual actions.
        @param log_context: String providing context to log messages
        @param raise_on_failure: If True, raise exception on action failures
        @return: True if all actions completed with success. False otherwise.
        """

        def upper_first(input_string):
            return input_string[0].upper() + input_string[1:] if len(input_string) > 0 else ''

        self.log_info(upper_first(log_context))
        result_list = []
        time_budget = Task.ACTION_TIMEOUT
        for action_worker, action_info in action_list:
            while True:
                action = ActionStatus.get(api, action_worker.uuid)
                if action is None:
                    self.log_warning('Failed to retrieve action status from vManage')
                    result_list.append(False)
                    break

                if action.is_completed:
                    result_list.append(action.is_successful)
                    if action_info is not None:
                        if action.is_successful:
                            self.log_info(f'Completed {action_info}')
                        else:
                            self.log_warning(f'Failed {action_info}: {action.activity_details}')

                    break

                time_budget -= Task.ACTION_INTERVAL
                if time_budget > 0:
                    self.log_info('Waiting...')
                    time.sleep(Task.ACTION_INTERVAL)
                else:
                    self.log_warning('Wait time limit expired')
                    result_list.append(False)
                    break

        result = all(result_list)
        if result:
            self.log_info(f'Completed {log_context}')
        elif raise_on_failure:
            raise WaitActionsException(f'Failed {log_context}')
        else:
            self.log_warning(f'Failed {log_context}')

        return result


class TaskException(Exception):
    """ Exception for Task errors """
    pass


class WaitActionsException(TaskException):
    """ Exception indicating failure in one or more actions being monitored """
    pass


def chopper(section_size: int):
    section = {}
    for _ in range(section_size):
        data = yield
        if data is None:
            break
        primary_key, secondary_key, item = data
        section.setdefault(primary_key, {}).setdefault(secondary_key, []).append(item)
    return section


def device_iter(api: Rest, match_name_regex: Optional[str] = None, match_reachable: bool = False,
                match_site_id: Optional[str] = None, match_system_ip: Optional[str] = None) -> Iterator[tuple]:
    """
    Return an iterator over device inventory, filtered by optional conditions.
    @param api: Instance of Rest API
    @param match_name_regex: Regular expression matching device host-name
    @param match_reachable: Boolean indicating whether to include reachable devices only
    @param match_site_id: When present, only include devices with provided site-id
    @param match_system_ip: If present, only include device with provided system-ip
    @return: Iterator of (<device-uuid>, <device-name>) tuples.
    """
    return (
        (uuid, name)
        for uuid, name, system_ip, site_id, reachability, *_ in Device.get_raise(api).extended_iter(default='-')
        if (
            (match_name_regex is None or regex_search(match_name_regex, name)) and
            (not match_reachable or reachability == 'reachable') and
            (match_site_id is None or site_id == match_site_id) and
            (match_system_ip is None or system_ip == match_system_ip)
        )
    )


def clean_dir(target_dir_name: str, max_saved: int = 99) -> Union[str, bool]:
    """
    Clean target_dir_name directory if it exists. If max_saved is non-zero and target_dir_name exists, move it to a new
    directory name in sequence.
    @param target_dir_name: str with the directory to be cleaned
    @param max_saved: int indicating the maximum instances to keep. If 0, target_dir_name is just deleted.
    """
    target_dir = Path(DATA_DIR, target_dir_name)
    if target_dir.exists():
        if max_saved > 0:
            save_seq = range(max_saved)
            for elem in save_seq:
                save_path = Path(DATA_DIR, f'{target_dir_name}_{elem + 1}')
                if elem == save_seq[-1]:
                    rmtree(save_path, ignore_errors=True)
                if not save_path.exists():
                    target_dir.rename(save_path)
                    return save_path.name
        else:
            rmtree(target_dir, ignore_errors=True)

    return False


def export_json(table_iter: Iterable[Table], filename: str) -> None:
    """
    Export a group (Iterable) of Tables as a JSON encoded file
    @param table_iter: Tables to export
    @param filename: Name for the export file
    """
    with open(filename, 'w') as export_file:
        data = [table.dict() for table in table_iter]
        json.dump(data, export_file, indent=2)



import argparse
from uuid import uuid4
from copy import deepcopy
from contextlib import suppress
from typing import Union, Optional, Tuple, List, Dict, Type, Callable
from pydantic import BaseModel, validator, ValidationError, Extra
import yaml
from cisco_sdwan.__version__ import __doc__ as title
from cisco_sdwan.base.rest_api import Rest
from cisco_sdwan.base.catalog import catalog_iter, CATALOG_TAG_ALL, ordered_tags
from cisco_sdwan.base.models_base import update_ids, ServerInfo, ConfigItem, ExtendedTemplate
from cisco_sdwan.base.models_vmanage import DeviceTemplate, DeviceTemplateAttached, DeviceTemplateValues
from cisco_sdwan.base.processor import StopProcessorException, ProcessorException
from cisco_sdwan.tasks.utils import (TaskOptions, TagOptions, existing_workdir_type, filename_type, ext_template_type,
                                     regex_type, existing_file_type)
from cisco_sdwan.tasks.common import clean_dir, Task, TaskException, regex_search
from cisco_sdwan.tasks.models import (const, TaskArgs, validate_workdir, validate_ext_template, validate_filename,
                                      validate_regex, validate_catalog_tag, validate_existing_file, validate_json)


class RecipeException(Exception):
    """ Exception indicating issues with the transform recipe """
    pass


class TransformRecipeNameTemplate(BaseModel, extra=Extra.forbid):
    regex: Optional[str] = None
    not_regex: Optional[str] = None
    name_regex: str

    _validate_regex = validator('regex', 'not_regex', allow_reuse=True)(validate_regex)
    _validate_name_regex = validator('name_regex', allow_reuse=True)(validate_ext_template)


class TransformRecipe(BaseModel, extra=Extra.forbid):
    tag: str
    name_template: Optional[TransformRecipeNameTemplate] = None
    name_map: Optional[Dict[str, str]] = None
    replace_source: bool = True

    # Validators
    _validate_tag = validator('tag', allow_reuse=True)(validate_catalog_tag)

    @validator('name_map',  always=True)
    def validate_name_map(cls, v, values):
        if v is None and values.get('name_template') is None:
            raise ValueError('At least one of "name_map" or "name_template" needs to be provided')
        return v

    def __init__(self, **kwargs):
        # Dummy init used so PyCharm type checker can recognize parameters
        super().__init__(**kwargs)

    @classmethod
    def parse_yaml(cls, filename: str):
        try:
            with open(filename) as yaml_file:
                recipe_dict = yaml.safe_load(yaml_file)
                return cls.parse_obj(recipe_dict)
        except FileNotFoundError as ex:
            raise RecipeException(f'Could not load recipe file: {ex}') from None
        except yaml.YAMLError as ex:
            raise RecipeException(f'Recipe file YAML syntax error: {ex}') from None


class Processor:
    def __init__(self, name: str, recipe: TransformRecipe):
        self.name = name
        self.recipe = recipe

    def match(self, name: str, tag: str) -> Union[str, None]:
        """
        Create a new name. If name_map is defined in the recipe and there is mapping for the current name, then use it.
        Otherwise, if name_regex is defined in the recipe, then use it to build the new name.
        """
        # Match tag
        if self.recipe.tag != CATALOG_TAG_ALL and self.recipe.tag != tag:
            return None

        # Match name_map
        new_name = self.recipe.name_map.get(name) if self.recipe.name_map is not None else None

        # Match regex / name_regex
        if new_name is None and self.recipe.name_template is not None:
            regex = self.recipe.name_template.regex or self.recipe.name_template.not_regex
            if regex is None or regex_search(regex, name, inverse=self.recipe.name_template.regex is None):
                new_name = ExtendedTemplate(self.recipe.name_template.name_regex)(name)

        return new_name

    def eval(self, config_obj: ConfigItem, new_name: str, new_id: str) -> Tuple[dict, List[str]]:
        new_payload = deepcopy(config_obj.data)
        trace_log: List[str] = []

        if config_obj.name_tag:
            new_payload[config_obj.name_tag] = new_name
        # In older releases, device templates did not have a templateId
        if config_obj.id_tag and config_obj.id_tag in new_payload:
            new_payload[config_obj.id_tag] = new_id

        # Reset attributes that would make this item read-Only
        for ro_tag in (config_obj.factory_default_tag, config_obj.readonly_tag):
            if new_payload.get(ro_tag, False):
                new_payload[ro_tag] = False
                trace_log.append(f'Resetting "{ro_tag}" flag to "False"')

        return new_payload, trace_log

    @property
    def replace_source(self):
        return self.recipe.replace_source


class AttachedProcessor(Processor):
    def eval(self, config_obj: ConfigItem, new_name: str, new_id: str) -> Tuple[dict, List[str]]:
        new_payload = deepcopy(config_obj.data)
        trace_log: List[str] = []

        return new_payload, trace_log


class ValuesProcessor(Processor):
    def eval(self, config_obj: ConfigItem, new_name: str, new_id: str) -> Tuple[dict, List[str]]:
        new_payload = deepcopy(config_obj.data)
        trace_log: List[str] = []

        return new_payload, trace_log


@TaskOptions.register('transform')
class TaskTransform(Task):
    @staticmethod
    def parser(task_args, target_address=None):
        task_parser = argparse.ArgumentParser(description=f'{title}\nTransform task:')
        task_parser.prog = f'{task_parser.prog} transform'
        task_parser.formatter_class = argparse.RawDescriptionHelpFormatter

        sub_tasks = task_parser.add_subparsers(title='transform options')
        sub_tasks.required = True

        rename_parser = sub_tasks.add_parser('rename', help='rename configuration items')
        rename_parser.set_defaults(recipe_handler=TaskTransform.rename_recipe)

        copy_parser = sub_tasks.add_parser('copy', help='copy configuration items')
        copy_parser.set_defaults(recipe_handler=TaskTransform.copy_recipe)

        recipe_parser = sub_tasks.add_parser('recipe', help='transform using custom recipe')
        recipe_parser.set_defaults(recipe_handler=TaskTransform.load_recipe)
        recipe_mutex = recipe_parser.add_mutually_exclusive_group(required=True)
        recipe_mutex.add_argument('--from-file', metavar='<filename>', type=existing_file_type,
                                  help='load recipe from YAML file')
        recipe_mutex.add_argument('--from-json', metavar='<json>',
                                  help='load recipe from JSON-formatted string')

        for sub_task in (rename_parser, copy_parser):
            sub_task.add_argument('tag', metavar='<tag>', type=TagOptions.tag,
                                  help='tag for selecting items to transform. Available tags: '
                                       f'{TagOptions.options()}. Special tag "{CATALOG_TAG_ALL}" selects all items.')
            mutex = sub_task.add_mutually_exclusive_group()
            mutex.add_argument('--regex', metavar='<regex>', type=regex_type,
                               help='regular expression selecting item names to transform')
            mutex.add_argument('--not-regex', metavar='<regex>', type=regex_type,
                               help='regular expression selecting item names NOT to transform')
            sub_task.add_argument('name_regex', metavar='<name-regex>', type=ext_template_type,
                                  help='name-regex used to transform an existing item name. Variable {name} is '
                                       'replaced with the original template name. Sections of the original template '
                                       'name can be selected using the {name <regex>} format. Where <regex> is a '
                                       'regular expression that must contain at least one capturing group. Capturing '
                                       'groups identify sections of the original name to keep.')

        for sub_task in (rename_parser, copy_parser, recipe_parser):
            sub_task.add_argument('output', metavar='<output-directory>', type=filename_type,
                                  help='directory to save transform result')
            sub_task.add_argument('--no-rollover', action='store_true',
                                  help='by default, if the output directory already exists it is renamed using a '
                                       'rolling naming scheme. This option disables this automatic rollover.')
            sub_task.add_argument('--workdir', metavar='<directory>', type=existing_workdir_type,
                                  help='transform will read from the specified directory instead of target vManage')

        return task_parser.parse_args(task_args)

    @staticmethod
    def is_api_required(parsed_args) -> bool:
        return parsed_args.workdir is None

    @staticmethod
    def _recipe_dict(parsed_args, replace_source: bool) -> dict:
        recipe_dict = {
            'tag': parsed_args.tag,
            'name_template': {
                'name_regex': parsed_args.name_regex
            },
            'replace_source': replace_source
        }
        if parsed_args.regex:
            recipe_dict['name_template']['regex'] = parsed_args.regex
        elif parsed_args.not_regex:
            recipe_dict['name_template']['not_regex'] = parsed_args.not_regex

        return recipe_dict

    @classmethod
    def rename_recipe(cls, parsed_args) -> TransformRecipe:
        return TransformRecipe(**cls._recipe_dict(parsed_args, replace_source=True))

    @classmethod
    def copy_recipe(cls, parsed_args) -> TransformRecipe:
        return TransformRecipe(**cls._recipe_dict(parsed_args, replace_source=False))

    @staticmethod
    def load_recipe(parsed_args) -> TransformRecipe:
        if parsed_args.from_file:
            return TransformRecipe.parse_yaml(parsed_args.from_file)
        else:
            return TransformRecipe.parse_raw(parsed_args.from_json)

    def runner(self, parsed_args, api: Optional[Rest] = None) -> Union[None, list]:
        source_info = f'Local workdir: "{parsed_args.workdir}"' if api is None else f'vManage URL: "{api.base_url}"'
        self.log_info(f'Transform task: {source_info} -> Local output dir: "{parsed_args.output}"')

        # Load processors
        try:
            recipe = parsed_args.recipe_handler(parsed_args)
            default_processor = Processor(name='default', recipe=recipe)
            loaded_processors = {
                DeviceTemplateAttached: AttachedProcessor(name='attached devices processor', recipe=recipe),
                DeviceTemplateValues: ValuesProcessor(name='values processor', recipe=recipe)
            }
        except (ValidationError, RecipeException) as ex:
            raise TaskException(f'Error loading transform recipe: {ex}') from None

        # Output directory must be empty for a new transform
        saved_output = clean_dir(parsed_args.output, max_saved=0 if parsed_args.no_rollover else 99)
        if saved_output:
            self.log_info(f'Previous output under "{parsed_args.output}" was saved as "{saved_output}"')

        # Identify backend and server version
        if api is None:
            backend = parsed_args.workdir
            local_info = ServerInfo.load(backend)
            server_version = local_info.server_version if local_info is not None else None
        else:
            backend = api
            server_version = backend.server_version

        if server_version is not None:
            if ServerInfo(server_version=server_version).save(parsed_args.output):
                self.log_info('Saved vManage server information')

        # Process items
        id_mapping: Dict[str, str] = {}  # {<old_id>: <new_id>}
        try:
            for tag in ordered_tags(CATALOG_TAG_ALL, reverse=True):
                self.log_info(f'Inspecting {tag} items')

                for _, info, index_cls, item_cls in catalog_iter(tag, version=server_version):
                    item_index = self.index_get(index_cls, backend)
                    if item_index is None:
                        self.log_debug(f'Skipped {info}, none found')
                        continue

                    processor = loaded_processors.get(item_cls, default_processor)
                    name_set = {item_name for item_id, item_name in item_index}
                    is_bad_name = False
                    export_list = []
                    id_hint_map = {item_name: item_id for item_id, item_name in item_index}

                    for item_id, item_name in item_index:
                        item = self.retrieve(item_cls, backend, item_id, item_name, item_index.need_extended_name)
                        if item is None:
                            self.log_error(f'Failed loading {info} {item_name}')
                            continue

                        self.log_debug(f'Evaluating {info} {item_name} with {processor.name} processor')
                        with suppress(StopProcessorException):
                            new_name = processor.match(item_name, tag)
                            if new_name is None:
                                raise StopProcessorException()

                            self.log_info(f'Matched {info} {item_name}')
                            if not item_cls.is_name_valid(new_name):
                                self.log_error(f'New {info} name is not valid: {new_name}')
                                is_bad_name = True
                                raise StopProcessorException()
                            if new_name in name_set:
                                self.log_error(f'New {info} name already exists: {new_name}')
                                is_bad_name = True
                                raise StopProcessorException()

                            new_id = item_id if processor.replace_source else str(uuid4())
                            new_payload = self.processor_eval(processor, item, new_name, new_id)
                            new_item = item_cls(update_ids(id_mapping, new_payload))

                            if isinstance(item, DeviceTemplate):
                                if item.devices_attached is not None:
                                    b_processor = loaded_processors.get(DeviceTemplateAttached, default_processor)
                                    new_item.devices_attached = DeviceTemplateAttached(
                                        self.processor_eval(b_processor, item.devices_attached, new_name, new_id)
                                    )
                                if item.attach_values is not None:
                                    b_processor = loaded_processors.get(DeviceTemplateValues, default_processor)
                                    new_item.attach_values = DeviceTemplateValues(
                                        self.processor_eval(b_processor, item.attach_values, new_name, new_id)
                                    )

                            if processor.replace_source:
                                self.log_info(f'Replacing {info}: {item_name} -> {new_name}')
                                item = new_item
                            else:
                                self.log_info(f'Adding {info}: {new_name}')
                                export_list.append(new_item)
                                id_mapping[item_id] = new_id

                            name_set.add(new_name)
                            id_hint_map[new_name] = new_id

                        export_list.append(item)

                    if is_bad_name:
                        raise TaskException(f'One or more {info} new names are invalid')

                    if not export_list:
                        self.log_debug(f'No {info} to export')
                        continue

                    x_item_index = index_cls.create(export_list, id_hint_map)
                    if x_item_index.save(parsed_args.output):
                        self.log_debug(f'Saved {info} index')

                    for x_item in export_list:
                        save_params = (parsed_args.output, x_item_index.need_extended_name, x_item.name,
                                       id_hint_map[x_item.name])
                        if x_item.save(*save_params):
                            self.log_debug(f'Saved {info} {x_item.name}')

                        if isinstance(x_item, DeviceTemplate):
                            if x_item.devices_attached is not None and x_item.devices_attached.save(*save_params):
                                self.log_debug(f'Saved {info} {x_item.name} attached devices')
                            if x_item.attach_values is not None and x_item.attach_values.save(*save_params):
                                self.log_debug(f'Saved {info} {x_item.name} values')

        except ProcessorException as ex:
            raise TaskException(f'Transform aborted: {ex}') from None

        return

    def retrieve(self, item_cls: Type[ConfigItem], backend: Union[Rest, str],
                 item_id: str, item_name: str, ext_name: bool) -> Union[ConfigItem, None]:

        item = self.item_get(item_cls, backend, item_id, item_name, ext_name)
        if item is None:
            return None

        if isinstance(item, DeviceTemplate):
            is_api = isinstance(backend, Rest)
            if is_api:
                devices_attached = DeviceTemplateAttached.get_raise(backend, item_id)
            else:
                # devices_attached will be None if there are no attachments (i.e. file is not present)
                devices_attached = DeviceTemplateAttached.load(backend, ext_name, item_name, item_id)

            if devices_attached is not None and not devices_attached.is_empty:
                if is_api:
                    uuid_iter = (uuid for uuid, _ in devices_attached)
                    attach_values = DeviceTemplateValues.get_values(backend, item_id, uuid_iter)
                else:
                    attach_values = DeviceTemplateValues.load(backend, ext_name, item_name, item_id)

                if attach_values is None:
                    self.log_error(f'Failed loading {item_name} values')
                else:
                    item.devices_attached = devices_attached
                    item.attach_values = attach_values

        return item

    def processor_eval(self, p: Processor, config_obj: ConfigItem, new_name: str, new_id: str) -> dict:
        new_payload, trace_log = p.eval(config_obj, new_name, new_id)
        for trace in trace_log:
            self.log_debug(f'Processor {p.name}, {new_name}: {trace}')

        return new_payload


class TransformArgs(TaskArgs):
    output: str
    workdir: Optional[str] = None
    no_rollover: bool = False

    # Validators
    _validate_filename = validator('output', allow_reuse=True)(validate_filename)
    _validate_workdir = validator('workdir', allow_reuse=True)(validate_workdir)


class TransformCopyArgs(TransformArgs):
    recipe_handler: Callable = const(TaskTransform.copy_recipe)
    tag: str
    regex: Optional[str] = None
    not_regex: Optional[str] = None
    name_regex: str

    # Validators
    _validate_tag = validator('tag', allow_reuse=True)(validate_catalog_tag)
    _validate_regex = validator('regex', 'not_regex', allow_reuse=True)(validate_regex)
    _validate_name = validator('name_regex', allow_reuse=True)(validate_ext_template)


class TransformRenameArgs(TransformCopyArgs):
    recipe_handler: Callable = const(TaskTransform.rename_recipe)


class TransformRecipeArgs(TransformArgs):
    recipe_handler: Callable = const(TaskTransform.load_recipe)
    from_file: Optional[str] = None
    from_json: Optional[str] = None

    # Validators
    _validate_existing_file = validator('from_file', allow_reuse=True)(validate_existing_file)
    _validate_json = validator('from_json', allow_reuse=True)(validate_json)



"""
 Sastre - Automation Tools for Cisco SD-WAN Powered by Viptela

 cisco_sdwan.tasks.implementation
 This module contains the implementation of user-facing tasks
"""
from ._backup import TaskBackup, BackupArgs
from ._restore import TaskRestore, RestoreArgs
from ._delete import TaskDelete, DeleteArgs
from ._migrate import TaskMigrate, MigrateArgs
from ._attach_detach import TaskAttach, TaskDetach, AttachVsmartArgs, AttachEdgeArgs, DetachVsmartArgs, DetachEdgeArgs
from ._transform import TaskTransform, TransformCopyArgs, TransformRenameArgs, TransformRecipeArgs

__all__ = [
    'TaskBackup',
    'TaskRestore',
    'TaskDelete',
    'TaskMigrate',
    'TaskAttach',
    'TaskDetach',
    'TaskTransform',
    'BackupArgs',
    'RestoreArgs',
    'DeleteArgs',
    'MigrateArgs',
    'AttachVsmartArgs',
    'AttachEdgeArgs',
    'DetachVsmartArgs',
    'DetachEdgeArgs',
    'TransformCopyArgs',
    'TransformRenameArgs',
    'TransformRecipeArgs'
]



"""
Sastre - Automation Tools for Cisco SD-WAN Powered by Viptela

"""
__copyright__ = "Copyright (c) 2019-2022 Cisco Systems, Inc. and/or its affiliates"
__version__   = "1.18.1"
__author__    = "Marcelo Reis"
__email__     = "sastre-support@cisco.com"
__url__       = "https://github.com/CiscoDevNet/sastre"



"""
  Sastre - Automation Tools for Cisco SD-WAN Powered by Viptela

  cisco_sdwan.cmd
  This module implements the command line top-level parser and task dispatcher
"""
import logging
import logging.config
import logging.handlers
import argparse
import json
import sys
from pathlib import Path
from requests.exceptions import ConnectionError, HTTPError
from .base.rest_api import Rest, RestAPIException
from .base.catalog import catalog_size, op_catalog_size
from .base.models_base import ModelException, SASTRE_ROOT_DIR
from .__version__ import __version__ as version
from .__version__ import __doc__ as title
from .tasks.utils import TaskOptions, EnvVar, non_empty_type, PromptArg
from .tasks.common import TaskException
from .tasks.implementation import *

# vManage REST API defaults
VMANAGE_PORT = '443'
REST_TIMEOUT = 300

# Default logging configuration - JSON formatted
# Setting level at chardet.charsetprober to prevent unwanted debug messages from requests module
LOGGING_CONFIG = '''
{
    "version": 1,
    "formatters": {
        "simple": {
            "format": "%(levelname)s: %(message)s"
        },
        "detailed": {
            "format": "%(asctime)s: %(name)s: %(levelname)s: %(message)s"
        }
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "level": "WARN",
            "formatter": "simple"
        },
        "file": {
            "class": "logging.handlers.RotatingFileHandler",
            "filename": "logs/sastre.log",
            "backupCount": 3,
            "maxBytes": 204800,
            "level": "DEBUG",
            "formatter": "detailed"
        }
    },
    "root": {
        "handlers": ["console", "file"],
        "level": "DEBUG"
    },
    "loggers": {
        "chardet.charsetprober": {
            "level": "INFO"
        }
    }
}
'''


def main():
    # Top-level cli parser
    cli_parser = argparse.ArgumentParser(description=title)
    cli_parser.add_argument('-a', '--address', metavar='<vmanage-ip>', action=EnvVar, required=False,
                            envvar='VMANAGE_IP', type=non_empty_type,
                            help='vManage IP address, can also be defined via VMANAGE_IP environment variable. '
                                 'If neither is provided user is prompted for the address.')
    cli_parser.add_argument('-u', '--user', metavar='<user>', action=EnvVar, required=False,
                            envvar='VMANAGE_USER', type=non_empty_type,
                            help='username, can also be defined via VMANAGE_USER environment variable. '
                                 'If neither is provided user is prompted for username.')
    cli_parser.add_argument('-p', '--password', metavar='<password>', action=EnvVar, required=False,
                            envvar='VMANAGE_PASSWORD', type=non_empty_type,
                            help='password, can also be defined via VMANAGE_PASSWORD environment variable. '
                                 ' If neither is provided user is prompted for password.')
    cli_parser.add_argument('--tenant', metavar='<tenant>', type=non_empty_type,
                            help='tenant name, when using provider accounts in multi-tenant deployments.')
    cli_parser.add_argument('--port', metavar='<port>', default=VMANAGE_PORT, action=EnvVar, envvar='VMANAGE_PORT',
                            help='vManage port number, can also be defined via VMANAGE_PORT environment variable '
                                 '(default: %(default)s)')
    cli_parser.add_argument('--timeout', metavar='<timeout>', type=int, default=REST_TIMEOUT,
                            help='REST API timeout (default: %(default)s)')
    cli_parser.add_argument('--verbose', action='store_true',
                            help='increase output verbosity')
    cli_parser.add_argument('--version', action='version',
                            version=f'Sastre Version {version}. Catalog: {catalog_size()} configuration items, '
                                    f'{op_catalog_size()} operational items.')
    cli_parser.add_argument('task', metavar='<task>', type=TaskOptions.task,
                            help=f'task to be performed ({TaskOptions.options()})')
    cli_parser.add_argument('task_args', metavar='<arguments>', nargs=argparse.REMAINDER,
                            help='task parameters, if any')
    cli_parser.set_defaults(prompt_arguments_api=[
        PromptArg('address', 'vManage address: '),
        PromptArg('user', 'vManage user: '),
        PromptArg('password', 'vManage password: ', secure_prompt=True)
    ])
    cli_args = cli_parser.parse_args()

    # Logging setup
    logging_config = json.loads(LOGGING_CONFIG)
    console_handler = logging_config.get('handlers', {}).get('console')
    if cli_args.verbose and console_handler is not None:
        console_handler['level'] = 'INFO'

    file_handler = logging_config.get('handlers', {}).get('file')
    if file_handler is not None:
        file_handler['filename'] = str(Path(SASTRE_ROOT_DIR, file_handler['filename']))
        Path(file_handler['filename']).parent.mkdir(parents=True, exist_ok=True)

    logging.config.dictConfig(logging_config)

    # Prepare task
    task = cli_args.task()
    target_address = cli_args.address
    parsed_task_args = task.parser(cli_args.task_args, target_address=target_address)
    is_api_required = task.is_api_required(parsed_task_args)

    # Evaluate whether user must be prompted for additional arguments
    prompt_args_list = getattr(cli_args, 'prompt_arguments', [])
    if is_api_required:
        prompt_args_list.extend(getattr(cli_args, 'prompt_arguments_api', []))
    try:
        for prompt_arg in prompt_args_list:
            if getattr(cli_args, prompt_arg.argument) is None:
                setattr(cli_args, prompt_arg.argument, prompt_arg())
    except KeyboardInterrupt:
        sys.exit(1)

    # Dispatch task
    try:
        if is_api_required:
            if target_address != cli_args.address:
                # Target address changed, re-run parser
                parsed_task_args = task.parser(cli_args.task_args, target_address=cli_args.address)

            base_url = f'https://{cli_args.address}{"" if cli_args.port == "443" else f":{cli_args.port}"}'
            with Rest(base_url, cli_args.user, cli_args.password, cli_args.tenant, timeout=cli_args.timeout) as api:
                # Dispatch to the appropriate task handler
                task_output = task.runner(parsed_task_args, api)

        else:
            # Dispatch to the appropriate task handler without api connection
            task_output = task.runner(parsed_task_args)

        # Display task output
        if task_output:
            print('\n\n'.join(str(entry) for entry in task_output))
        # Display dryrun report if console logging is disabled (i.e. not in verbose mode)
        if not cli_args.verbose and task.is_dryrun:
            print(str(task.dryrun_report))

        task.log_info(f'Task completed {task.outcome("successfully", "with caveats: {tally}")}')
    except (RestAPIException, ConnectionError, HTTPError, FileNotFoundError, ModelException, TaskException) as ex:
        logging.getLogger(__name__).critical(ex)
    except KeyboardInterrupt:
        logging.getLogger(__name__).critical('Interrupted by user')



from copy import deepcopy
from collections.abc import Hashable
from cisco_sdwan.base.processor import Operation, Processor, ProcessorException
from . import module_dir, factory_cedge_aaa, factory_cedge_global

_operations = {}  # {<operation_key>: Operation ...}


def register(operation_key, *param_keys):
    """
    Decorator used for registering operation handlers.
    @param operation_key: Operation key as used in the recipe file
    @param param_keys: Parameter keys from recipe file to be supplied to operation handler call.
    @return: decorator
    """

    def decorator(operation_fn):
        _operations[operation_key] = Operation(operation_fn, param_keys)
        return operation_fn

    return decorator


@register('replace', 'fieldHierarchyList', 'mappings')
def op_replace(template_data, field_hierarchy_list, mappings):
    """
    Search for field_hierarchy_list elements in template_data and perform replacements according to mappings
    """
    op_trace = []

    def replace(json_obj, search_list):
        if len(search_list) == 0:
            return

        if isinstance(json_obj, dict):
            for key, value in json_obj.items():
                if key == search_list[0]:
                    if len(search_list) > 1:
                        replace(value, search_list[1:])
                    elif isinstance(value, Hashable):
                        new_val = mappings.get(value)
                        if new_val is not None:
                            op_trace.append('Template {name}, updated {path}: '
                                            '{from_val} -> {to_val}'.format(name=template_data['templateName'],
                                                                            path='/'.join(field_hierarchy_list),
                                                                            from_val=value,
                                                                            to_val=new_val))
                            json_obj[key] = new_val
                else:
                    replace(value, search_list)

        elif isinstance(json_obj, list):
            for elem in json_obj:
                replace(elem, search_list)

    replace(template_data, field_hierarchy_list)

    return op_trace


@register('remove', 'fieldHierarchyList', 'value')
def op_remove(template_data, field_hierarchy_list, value):
    """
    Search for field_hierarchy_list elements in template_data and remove entries matching value.
    Can only remove list entries, it does not remove dictionary entries.
    """
    op_trace = []

    def match_value(elem, field):
        if isinstance(elem, dict) and elem.get(field) == value:
            op_trace.append('Template {name}, removed {path}: {value}'.format(
                name=template_data['templateName'],
                path='/'.join(field_hierarchy_list),
                value=value))
            return True
        else:
            return False

    def remove(json_obj, search_list):
        if len(search_list) == 0:
            return

        if isinstance(json_obj, dict):
            for k, v in json_obj.items():
                if k == search_list[0]:
                    if len(search_list) > 1:
                        remove(v, search_list[1:])
                else:
                    remove(v, search_list)

        elif isinstance(json_obj, list):
            if len(search_list) == 1:
                json_obj[:] = [elem for elem in json_obj if not match_value(elem, search_list[0])]

            for elem in json_obj:
                remove(elem, search_list)

    remove(template_data, field_hierarchy_list)

    return op_trace


def add_template(payload_dict: dict, template_type: str, template_id: str) -> bool:
    general_templates = payload_dict.get('generalTemplates')
    if general_templates is None:
        return False

    general_templates.append(
        {
            "templateType": template_type,
            "templateId": template_id
        }
    )
    return True


class DeviceProcessor(Processor):
    recipe_file = module_dir().joinpath('device_template_recipes.json')

    mandatory_keys = {
        'squashedVmanageVersions': {},
        'tovManageVersion': {},
        'listOfTasks': {
            'operation': {},
            'fieldHierarchyList': {}
        }
    }

    def __init__(self, data, from_version, to_version):
        # Extract recipes
        matched_recipes = [
            recipe for recipe in data
            if from_version in recipe['squashedVmanageVersions'] and to_version == recipe['tovManageVersion']
        ]
        if not matched_recipes:
            raise ProcessorException('No recipe available to migrate from {from_version} to {to_version}'.format(
                from_version=from_version, to_version=to_version))

        super().__init__(matched_recipes)

    def is_in_scope(self, device_template, **kwargs):
        return device_template.is_cedge and not device_template.is_type_cli

    def eval(self, device_template, new_name, new_id):
        migrated_payload = deepcopy(device_template.data)
        trace_log = []

        old_name = device_template.name

        if not device_template.contains_template('cedge_aaa'):
            if not add_template(migrated_payload, 'cedge_aaa', factory_cedge_aaa.uuid):
                raise ProcessorException(f'Unable to attach cedge_aaa factory default to {old_name}')
            trace_log.append(f'Attached cedge_aaa factory default to {old_name}')

        if not device_template.contains_template('cedge_global'):
            if not add_template(migrated_payload, 'cedge_global', factory_cedge_global.uuid):
                raise ProcessorException(f'Unable to attach cedge_global factory default to {old_name}')
            trace_log.append(f'Attached cedge_global factory default to {old_name}')

        for recipe in self.data:
            for task in recipe['listOfTasks']:
                op = _operations.get(task['operation'])
                if op is None:
                    trace_log.append('Operation {task_op} is not supported, skipping'.format(task_op=task['operation']))
                    continue

                param_values = (task.get(param_key) for param_key in op.param_keys)
                op_trace = op.handler_fn(migrated_payload, *param_values)
                trace_log.extend(op_trace)

        if 'templateClass' in migrated_payload:
            migrated_payload['templateClass'] = 'cedge'
        else:
            trace_log.append(f'No templateClass in {old_name}')

        migrated_payload['templateName'] = new_name

        return migrated_payload, trace_log



import json
from collections import namedtuple
from copy import deepcopy

Operation = namedtuple('Operation', ['handler_fn', 'param_keys'])


class Processor:
    recipe_file = None
    mandatory_keys = {
    }

    def __init__(self, data, **kwargs):
        self.data = data

    @classmethod
    def validate_recipe_data(cls, recipe_data):
        violations_list = []

        def validate(json_obj, violations, mandatory_keys_dict, bread_crumbs):
            if isinstance(json_obj, dict):
                for key in mandatory_keys_dict:
                    if key not in json_obj:
                        violations.append('{key} @ {crumbs}'.format(key=key, crumbs='/'.join(map(str, bread_crumbs))))
                    else:
                        validate(json_obj[key], violations, mandatory_keys_dict[key], bread_crumbs + [key])

            elif isinstance(json_obj, list):
                for index, elem in enumerate(json_obj):
                    validate(elem, violations, mandatory_keys_dict, bread_crumbs + [index])

            return violations

        return validate(recipe_data, violations_list, cls.mandatory_keys, [])

    @classmethod
    def load(cls, **kwargs):
        try:
            with open(cls.recipe_file, 'r') as read_f:
                data = json.load(read_f)
            assert isinstance(data, list)
        except FileNotFoundError:
            raise ProcessorException(f'Migration recipe file not found: {cls.recipe_file}') from None
        except json.decoder.JSONDecodeError as ex:
            raise ProcessorException(f'Invalid JSON in recipe file: {cls.recipe_file}: {ex}') from None
        except AssertionError:
            raise ProcessorException(f'Invalid recipe file: {cls.recipe_file}: Top level must be a list') from None

        # Enforce mandatory_keys
        violations = cls.validate_recipe_data(data)
        if violations:
            raise ProcessorException(
                f"Invalid recipe file: {cls.recipe_file}: Missing mandatory keys: {', '.join(violations)}"
            )

        return cls(data, **kwargs)

    def is_in_scope(self, template_obj, **kwargs):
        raise NotImplemented('Method not implemented')

    def eval(self, template_obj, new_name, new_id):
        migrated_payload = deepcopy(template_obj.data)
        trace_log = []

        return migrated_payload, trace_log

    def replace_original(self):
        """ By default, items generated by the processor would replace the original item """
        return True


class StopProcessorException(Exception):
    """ Exception indicating need to stop further processing of this item """
    pass


class ProcessorException(Exception):
    """ Exception for Migration Processor errors """
    pass



import argparse
from typing import Union, Optional
from pydantic import validator
from cisco_sdwan.__version__ import __doc__ as title
from cisco_sdwan.base.rest_api import Rest, RestAPIException
from cisco_sdwan.base.catalog import catalog_iter, CATALOG_TAG_ALL, ordered_tags
from cisco_sdwan.base.models_vmanage import DeviceTemplateIndex
from cisco_sdwan.tasks.utils import TaskOptions, TagOptions, regex_type
from cisco_sdwan.tasks.common import regex_search, Task, WaitActionsException
from cisco_sdwan.tasks.models import TaskArgs, validate_regex, validate_catalog_tag


@TaskOptions.register('delete')
class TaskDelete(Task):
    @staticmethod
    def parser(task_args, target_address=None):
        task_parser = argparse.ArgumentParser(description=f'{title}\nDelete task:')
        task_parser.prog = f'{task_parser.prog} delete'
        task_parser.formatter_class = argparse.RawDescriptionHelpFormatter

        mutex = task_parser.add_mutually_exclusive_group()
        mutex.add_argument('--regex', metavar='<regex>', type=regex_type,
                           help='regular expression matching item names to delete, within selected tags.')
        mutex.add_argument('--not-regex', metavar='<regex>', type=regex_type,
                           help='regular expression matching item names NOT to delete, within selected tags.')
        task_parser.add_argument('--dryrun', action='store_true',
                                 help='dry-run mode. Items matched for removal are listed but not deleted.')
        task_parser.add_argument('--detach', action='store_true',
                                 help='USE WITH CAUTION! Detach devices from templates and deactivate vSmart policy '
                                      'before deleting items. This allows deleting items that are associated with '
                                      'attached templates and active policies.')
        task_parser.add_argument('tag', metavar='<tag>', type=TagOptions.tag,
                                 help='tag for selecting items to be deleted. Available tags: '
                                      f'{TagOptions.options()}. Special tag "{CATALOG_TAG_ALL}" selects all items.')
        return task_parser.parse_args(task_args)

    def runner(self, parsed_args, api: Optional[Rest] = None) -> Union[None, list]:
        self.is_dryrun = parsed_args.dryrun
        self.log_info(f'Delete task: vManage URL: "{api.base_url}"')

        if parsed_args.detach:
            try:
                template_index = DeviceTemplateIndex.get_raise(api)
                # Detach WAN Edge templates
                reqs = self.detach(api, template_index.filtered_iter(DeviceTemplateIndex.is_not_vsmart),
                                   log_context='detaching WAN Edges')
                if reqs:
                    self.log_debug(f'Detach requests processed: {reqs}')
                else:
                    self.log_info('No WAN Edge attached')
                # Deactivate vSmart policy
                if not self.is_dryrun:
                    action_list = self.deactivate_policy(api)
                    if len(action_list) == 0:
                        self.log_info('No vSmart policy activated')
                    else:
                        self.wait_actions(api, action_list, 'deactivating vSmart policy', raise_on_failure=True)
                # Detach vSmart template
                reqs = self.detach(api, template_index.filtered_iter(DeviceTemplateIndex.is_vsmart),
                                   log_context='detaching vSmarts')
                if reqs:
                    self.log_debug(f'Detach requests processed: {reqs}')
                else:
                    self.log_info('No vSmart attached')
            except (RestAPIException, WaitActionsException) as ex:
                self.log_critical(f'Detach failed: {ex}')
                return

        for tag in ordered_tags(parsed_args.tag, parsed_args.tag != CATALOG_TAG_ALL):
            self.log_info(f'Inspecting {tag} items', dryrun=False)
            regex = parsed_args.regex or parsed_args.not_regex
            matched_item_iter = (
                (item_name, item_id, item_cls, info)
                for _, info, index, item_cls in self.index_iter(api, catalog_iter(tag, version=api.server_version))
                for item_id, item_name in index
                if regex is None or regex_search(regex, item_name, inverse=parsed_args.regex is None)
            )
            for item_name, item_id, item_cls, info in matched_item_iter:
                item = item_cls.get(api, item_id)
                if item is None:
                    self.log_warning(f'Failed retrieving {info} {item_name}')
                    continue
                if item.is_readonly or item.is_system:
                    self.log_debug(f'Skipped {"read-only" if item.is_readonly else "system"} {info} {item_name}')
                    continue
                if self.is_dryrun:
                    self.log_info(f'Delete {info} {item_name}')
                    continue

                if api.delete(item_cls.api_path.delete, item_id):
                    self.log_info(f'Done: Delete {info} {item_name}')
                else:
                    self.log_warning(f'Failed deleting {info} {item_name}')

        return


class DeleteArgs(TaskArgs):
    regex: Optional[str] = None
    not_regex: Optional[str] = None
    dryrun: bool = False
    detach: bool = False
    tag: str

    # Validators
    _validate_regex = validator('regex', 'not_regex', allow_reuse=True)(validate_regex)
    _validate_tag = validator('tag', allow_reuse=True)(validate_catalog_tag)



import argparse
from typing import Union, Optional
from pydantic import validator
from cisco_sdwan.__version__ import __doc__ as title
from cisco_sdwan.base.rest_api import Rest, RestAPIException, is_version_newer
from cisco_sdwan.base.catalog import catalog_iter, CATALOG_TAG_ALL, ordered_tags
from cisco_sdwan.base.models_base import UpdateEval, ServerInfo
from cisco_sdwan.base.models_vmanage import (DeviceTemplateIndex, PolicyVsmartIndex, EdgeInventory, ControlInventory,
                                             CheckVBond)
from cisco_sdwan.tasks.utils import TaskOptions, TagOptions, existing_workdir_type, regex_type, default_workdir
from cisco_sdwan.tasks.common import regex_search, Task, WaitActionsException
from cisco_sdwan.tasks.models import TaskArgs, validate_workdir, validate_regex, validate_catalog_tag


@TaskOptions.register('restore')
class TaskRestore(Task):
    @staticmethod
    def parser(task_args, target_address=None):
        task_parser = argparse.ArgumentParser(description=f'{title}\nRestore task:')
        task_parser.prog = f'{task_parser.prog} restore'
        task_parser.formatter_class = argparse.RawDescriptionHelpFormatter

        task_parser.add_argument('--workdir', metavar='<directory>', type=existing_workdir_type,
                                 default=default_workdir(target_address),
                                 help='restore source (default: %(default)s)')
        mutex = task_parser.add_mutually_exclusive_group()
        mutex.add_argument('--regex', metavar='<regex>', type=regex_type,
                           help='regular expression matching item names to restore, within selected tags.')
        mutex.add_argument('--not-regex', metavar='<regex>', type=regex_type,
                           help='regular expression matching item names NOT to restore, within selected tags.')
        task_parser.add_argument('--dryrun', action='store_true',
                                 help='dry-run mode. Items to be restored are listed but not pushed to vManage.')
        task_parser.add_argument('--attach', action='store_true',
                                 help='attach devices to templates and activate vSmart policy after restoring items')
        task_parser.add_argument('--update', action='store_true',
                                 help='update vManage items that have the same name but different content as the '
                                      'corresponding item in workdir. Without this option, such items are skipped '
                                      'from restore.')
        task_parser.add_argument('tag', metavar='<tag>', type=TagOptions.tag,
                                 help='tag for selecting items to be restored. Items that are dependencies of the '
                                      'specified tag are automatically included. Available tags: '
                                      f'{TagOptions.options()}. Special tag "{CATALOG_TAG_ALL}" selects all items.')
        return task_parser.parse_args(task_args)

    def runner(self, parsed_args, api: Optional[Rest] = None) -> Union[None, list]:
        def load_items(index, item_cls):
            item_iter = (
                (item_id, item_cls.load(parsed_args.workdir, index.need_extended_name, item_name, item_id))
                for item_id, item_name in index
            )
            return ((item_id, item_obj) for item_id, item_obj in item_iter if item_obj is not None)

        self.is_dryrun = parsed_args.dryrun
        self.log_info(f'Restore task: Local workdir: "{parsed_args.workdir}" -> vManage URL: "{api.base_url}"')

        local_info = ServerInfo.load(parsed_args.workdir)
        # Server info file may not be present (e.g. backup from older Sastre releases)
        if local_info is not None and is_version_newer(api.server_version, local_info.server_version):
            self.log_warning(f'Target vManage release ({api.server_version}) is older than the release used in backup '
                             f'({local_info.server_version}). Items may fail to restore due to incompatibilities.')

        is_vbond_set = self.is_vbond_configured(api)

        self.log_info('Loading existing items from target vManage', dryrun=False)
        target_all_items_map = {
            hash(type(index)): {item_name: item_id for item_id, item_name in index}
            for _, _, index, item_cls in self.index_iter(api, catalog_iter(CATALOG_TAG_ALL, version=api.server_version))
        }

        self.log_info('Identifying items to be pushed', dryrun=False)
        id_mapping = {}  # {<old_id>: <new_id>}, used to replace old (saved) item ids with new (target) ids
        restore_list = []  # [ (<info>, <index_cls>, [(<item_id>, <item>, <id_on_target>), ...]), ...]
        dependency_set = set()  # {<item_id>, ...}
        match_set = set()  # {<item_id>, ...}
        for tag in ordered_tags(parsed_args.tag):
            if tag == 'template_device' and not is_vbond_set:
                self.log_warning(f'Will skip {tag} items because vBond is not configured. '
                                 'On vManage, Administration > Settings > vBond.')
                continue

            self.log_info(f'Inspecting {tag} items', dryrun=False)
            tag_iter = (
                (info, index, load_items(index, item_cls))
                for _, info, index, item_cls in self.index_iter(parsed_args.workdir,
                                                                catalog_iter(tag, version=api.server_version))
            )
            for info, index, loaded_items_iter in tag_iter:
                target_item_map = target_all_items_map.get(hash(type(index)))
                if target_item_map is None:
                    # Logging at warning level because the backup files did have this item
                    self.log_warning(f'Will skip {info}, item not supported by target vManage')
                    continue

                restore_item_list = []
                for item_id, item in loaded_items_iter:
                    target_id = target_item_map.get(item.name)
                    if target_id is not None:
                        # Item already exists on target vManage, record item id from target
                        if item_id != target_id:
                            id_mapping[item_id] = target_id

                        if not parsed_args.update:
                            # Existing item on target vManage will be used, i.e. will not update it
                            self.log_debug(f'Will skip {info} {item.name}, item already on target vManage')
                            continue

                    if item.is_readonly:
                        self.log_debug(f'Will skip read-only {info} {item.name}')
                        continue

                    regex = parsed_args.regex or parsed_args.not_regex
                    item_matches = (
                            (parsed_args.tag == CATALOG_TAG_ALL or parsed_args.tag == tag) and
                            (regex is None or regex_search(regex, item.name, inverse=parsed_args.regex is None))
                    )
                    if item_matches:
                        match_set.add(item_id)
                    if item_matches or item_id in dependency_set:
                        # A target_id that is not None signals a put operation, as opposed to post.
                        # target_id will be None unless --update is specified and item name is on target
                        restore_item_list.append((item_id, item, target_id))
                        dependency_set.update(item.id_references_set)

                if len(restore_item_list) > 0:
                    restore_list.append((info, index, restore_item_list))

        if len(restore_list) > 0:
            self.log_info('Pushing items to vManage', dryrun=False)
            # Items were added to restore_list following ordered_tags() order (i.e. higher level items before lower
            # level items). The reverse order needs to be followed on restore.
            for info, index, restore_item_list in reversed(restore_list):
                pushed_item_dict = {}
                for item_id, item, target_id in restore_item_list:
                    op_info = 'Create' if target_id is None else 'Update'
                    reason = ' (dependency)' if item_id in dependency_set - match_set else ''

                    try:
                        if target_id is None:
                            # Create new item
                            if self.is_dryrun:
                                self.log_info(f'{op_info} {info} {item.name}{reason}')
                                continue
                            # Not using item id returned from post because post can return empty (e.g. local policies)
                            api.post(item.post_data(id_mapping), item.api_path.post)
                            pushed_item_dict[item.name] = item_id
                        else:
                            # Update existing item
                            update_data = item.put_data(id_mapping)
                            if item.get_raise(api, target_id).is_equal(update_data):
                                self.log_debug(f'{op_info} skipped (no diffs) {info} {item.name}')
                                continue

                            if self.is_dryrun:
                                self.log_info(f'{op_info} {info} {item.name}{reason}')
                                continue

                            put_eval = UpdateEval(api.put(update_data, item.api_path.put, target_id))
                            if put_eval.need_reattach:
                                if put_eval.is_master:
                                    self.log_info(f'Updating {info} {item.name} requires reattach')
                                    attach_data = self.reattach_template_data(api, [(item.name, target_id)])
                                else:
                                    self.log_info(
                                        f'Updating {info} {item.name} requires reattach of affected templates'
                                    )
                                    target_templates = {item_id: item_name
                                                        for item_id, item_name in DeviceTemplateIndex.get_raise(api)}
                                    templates_iter = (
                                        (target_templates[tgt_id], tgt_id)
                                        for tgt_id in put_eval.templates_affected_iter()
                                    )
                                    attach_data = self.reattach_template_data(api, templates_iter)

                                # All re-attachments need to be done in a single request, thus 9999 for chunk_size
                                reqs = self.attach(api, *attach_data, chunk_size=9999,
                                                   log_context='reattaching templates')
                                self.log_debug(f'Attach requests processed: {reqs}')
                            elif put_eval.need_reactivate:
                                self.log_info(f'Updating {info} {item.name} requires vSmart policy reactivate')
                                action_list = self.activate_policy(
                                    api, *PolicyVsmartIndex.get_raise(api).active_policy, is_edited=True
                                )
                                self.wait_actions(api, action_list, 'reactivating vSmart policy', raise_on_failure=True)
                    except (RestAPIException, WaitActionsException) as ex:
                        self.log_error(f'Failed {op_info} {info} {item.name}{reason}: {ex}')
                    else:
                        self.log_info(f'Done: {op_info} {info} {item.name}{reason}')

                # Read new ids from target and update id_mapping
                try:
                    new_target_item_map = {item_name: item_id for item_id, item_name in index.get_raise(api)}
                    for item_name, old_item_id in pushed_item_dict.items():
                        id_mapping[old_item_id] = new_target_item_map[item_name]
                except RestAPIException as ex:
                    self.log_critical(f'Failed retrieving {info}: {ex}')
                    break
        else:
            self.log_info('No items to push')

        if parsed_args.attach:
            try:
                target_templates = {item_name: item_id for item_id, item_name in DeviceTemplateIndex.get_raise(api)}
                target_policies = {item_name: item_id for item_id, item_name in PolicyVsmartIndex.get_raise(api)}
                saved_template_index = DeviceTemplateIndex.load(parsed_args.workdir, raise_not_found=True)

                # Attach WAN Edge templates
                edge_templates_iter = (
                    (saved_name, saved_id, target_templates.get(saved_name))
                    for saved_id, saved_name in saved_template_index.filtered_iter(DeviceTemplateIndex.is_not_vsmart)
                )
                attach_data = self.attach_template_data(
                    api, parsed_args.workdir, saved_template_index.need_extended_name, edge_templates_iter,
                    target_uuid_set={uuid for uuid, _ in EdgeInventory.get_raise(api)}
                )
                reqs = self.attach(api, *attach_data, log_context='attaching WAN Edges')
                if reqs:
                    self.log_debug(f'Attach requests processed: {reqs}')
                else:
                    self.log_info('No WAN Edge attachments needed')

                # Attach vSmart template
                vsmart_templates_iter = (
                    (saved_name, saved_id, target_templates.get(saved_name))
                    for saved_id, saved_name in saved_template_index.filtered_iter(DeviceTemplateIndex.is_vsmart)
                )
                vsmart_set = {
                    uuid for uuid, _ in ControlInventory.get_raise(api).filtered_iter(ControlInventory.is_vsmart)
                }
                attach_data = self.attach_template_data(api, parsed_args.workdir,
                                                        saved_template_index.need_extended_name, vsmart_templates_iter,
                                                        target_uuid_set=vsmart_set)
                reqs = self.attach(api, *attach_data, log_context="attaching vSmarts")
                if reqs:
                    self.log_debug(f'Attach requests processed: {reqs}')
                else:
                    self.log_info('No vSmart attachments needed')

                # Activate vSmart policy
                if not self.is_dryrun:
                    _, policy_name = PolicyVsmartIndex.load(parsed_args.workdir, raise_not_found=True).active_policy
                    action_list = self.activate_policy(api, target_policies.get(policy_name), policy_name)
                    if len(action_list) == 0:
                        self.log_info('No vSmart policy to activate')
                    else:
                        self.wait_actions(api, action_list, 'activating vSmart policy', raise_on_failure=True)
            except (RestAPIException, FileNotFoundError, WaitActionsException) as ex:
                self.log_critical(f'Attach failed: {ex}')

        return

    def is_vbond_configured(self, api: Rest) -> bool:
        if api.is_multi_tenant and not api.is_provider:
            # Cannot explicitly check vBond configuration with tenant account, assume it is configured
            return True

        check_vbond = CheckVBond.get(api)
        if check_vbond is None:
            self.log_warning('Failed retrieving vBond configuration status.')
            return False

        return check_vbond.is_configured


class RestoreArgs(TaskArgs):
    workdir: str
    regex: Optional[str] = None
    not_regex: Optional[str] = None
    dryrun: bool = False
    attach: bool = False
    update: bool = False
    tag: str

    # Validators
    _validate_workdir = validator('workdir', allow_reuse=True)(validate_workdir)
    _validate_regex = validator('regex', 'not_regex', allow_reuse=True)(validate_regex)
    _validate_tag = validator('tag', allow_reuse=True)(validate_catalog_tag)



import argparse
from functools import partial
from typing import Union, Optional, Callable
from pydantic import validator, conint
from cisco_sdwan.__version__ import __doc__ as title
from cisco_sdwan.base.rest_api import Rest, RestAPIException
from cisco_sdwan.base.models_vmanage import DeviceTemplateIndex, EdgeInventory, ControlInventory
from cisco_sdwan.tasks.utils import (TaskOptions, existing_workdir_type, regex_type, default_workdir, ipv4_type,
                                     site_id_type, int_type)
from cisco_sdwan.tasks.common import regex_search, Task, WaitActionsException, device_iter
from cisco_sdwan.tasks.models import TaskArgs, const, validate_regex, validate_workdir, validate_site_id, validate_ipv4

# Default number of devices to include per attach/detach request. The value of 200 was adopted because it is what was
# validated in the lab
DEFAULT_BATCH_SIZE = 200


@TaskOptions.register('attach')
class TaskAttach(Task):
    @staticmethod
    def parser(task_args, target_address=None):
        task_parser = argparse.ArgumentParser(description=f'{title}\nAttach templates task:')
        task_parser.prog = f'{task_parser.prog} attach'
        task_parser.formatter_class = argparse.RawDescriptionHelpFormatter

        sub_tasks = task_parser.add_subparsers(title='attach options')
        sub_tasks.required = True

        edge_parser = sub_tasks.add_parser('edge', help='attach templates to WAN edges')
        edge_parser.set_defaults(template_filter=DeviceTemplateIndex.is_not_vsmart,
                                 device_set=TaskAttach.edge_set,
                                 set_title="WAN Edges")

        vsmart_parser = sub_tasks.add_parser('vsmart', help='attach template to vSmarts')
        vsmart_parser.set_defaults(template_filter=DeviceTemplateIndex.is_vsmart,
                                   device_set=TaskAttach.vsmart_set,
                                   set_title="vSmarts")

        # Parameters common to all sub-tasks
        for sub_task in (edge_parser, vsmart_parser):
            sub_task.add_argument('--workdir', metavar='<directory>', type=existing_workdir_type,
                                  default=default_workdir(target_address), help='attach source (default: %(default)s)')
            sub_task.add_argument('--templates', metavar='<regex>', type=regex_type,
                                  help='regular expression selecting templates to attach. Match on template name.')
            sub_task.add_argument('--devices', metavar='<regex>', type=regex_type,
                                  help='regular expression selecting devices to attach. Match on device name.')
            sub_task.add_argument('--reachable', action='store_true', help='select reachable devices only')
            sub_task.add_argument('--site', metavar='<id>', type=site_id_type, help='select devices with site ID')
            sub_task.add_argument('--system-ip', metavar='<ipv4>', type=ipv4_type, help='select device with system IP')
            sub_task.add_argument('--dryrun', action='store_true',
                                  help='dry-run mode. Attach operations are listed but not is pushed to vManage.')
            sub_task.add_argument('--batch', metavar='<size>', type=partial(int_type, 1, 9999),
                                  default=DEFAULT_BATCH_SIZE,
                                  help='maximum number of devices to include per vManage attach request '
                                       '(default: %(default)s)')

        return task_parser.parse_args(task_args)

    @staticmethod
    def edge_set(api: Rest) -> set:
        return {uuid for uuid, _ in EdgeInventory.get_raise(api)}

    @staticmethod
    def vsmart_set(api: Rest) -> set:
        return {uuid for uuid, _ in ControlInventory.get_raise(api).filtered_iter(ControlInventory.is_vsmart)}

    def runner(self, parsed_args, api: Optional[Rest] = None) -> Union[None, list]:
        self.is_dryrun = parsed_args.dryrun
        self.log_info(f'Attach templates task: Local workdir: "{parsed_args.workdir}" -> vManage URL: "{api.base_url}"')

        try:
            uuid_set = (
                    parsed_args.device_set(api) &
                    set(uuid for uuid, _ in
                        device_iter(api, parsed_args.devices, parsed_args.reachable, parsed_args.site,
                                    parsed_args.system_ip))
            )
            target_templates = {item_name: item_id for item_id, item_name in DeviceTemplateIndex.get_raise(api)}
            saved_template_index = DeviceTemplateIndex.load(parsed_args.workdir, raise_not_found=True)
            matched_templates = (
                (saved_name, saved_id, target_templates.get(saved_name))
                for saved_id, saved_name in saved_template_index.filtered_iter(parsed_args.template_filter)
                if parsed_args.templates is None or regex_search(parsed_args.templates, saved_name)
            )

            attach_data = self.attach_template_data(api, parsed_args.workdir, saved_template_index.need_extended_name,
                                                    matched_templates, target_uuid_set=uuid_set)
            reqs = self.attach(api, *attach_data, chunk_size=parsed_args.batch,
                               log_context=f"attaching {parsed_args.set_title}")
            if reqs:
                self.log_debug(f'Attach requests processed: {reqs}')
            else:
                self.log_info(f'No {parsed_args.set_title} attachments to process')

        except (RestAPIException, FileNotFoundError, WaitActionsException) as ex:
            self.log_critical(f'Attach failed: {ex}')

        return


class AttachArgs(TaskArgs):
    templates: Optional[str] = None
    devices: Optional[str] = None
    site: Optional[str] = None
    system_ip: Optional[str] = None
    reachable: bool = False
    dryrun: bool = False
    batch: conint(ge=1, lt=9999) = DEFAULT_BATCH_SIZE
    workdir: str

    # Validators
    _validate_regex = validator('templates', 'devices', allow_reuse=True)(validate_regex)
    _validate_site_id = validator('site', allow_reuse=True)(validate_site_id)
    _validate_ipv4 = validator('system_ip', allow_reuse=True)(validate_ipv4)
    _validate_workdir = validator('workdir', allow_reuse=True)(validate_workdir)


class AttachVsmartArgs(AttachArgs):
    template_filter: Callable = const(DeviceTemplateIndex.is_vsmart)
    device_set: Callable = const(TaskAttach.vsmart_set)
    set_title: str = const('vSmarts')


class AttachEdgeArgs(AttachArgs):
    template_filter: Callable = const(DeviceTemplateIndex.is_not_vsmart)
    device_set: Callable = const(TaskAttach.edge_set)
    set_title: str = const('WAN Edges')


@TaskOptions.register('detach')
class TaskDetach(Task):
    @staticmethod
    def parser(task_args, target_address=None):
        task_parser = argparse.ArgumentParser(description=f'{title}\nDetach templates task:')
        task_parser.prog = f'{task_parser.prog} detach'
        task_parser.formatter_class = argparse.RawDescriptionHelpFormatter

        sub_tasks = task_parser.add_subparsers(title='detach options')
        sub_tasks.required = True

        edge_parser = sub_tasks.add_parser('edge', help='detach templates from WAN edges')
        edge_parser.set_defaults(template_filter=DeviceTemplateIndex.is_not_vsmart,
                                 set_title="WAN Edges")

        vsmart_parser = sub_tasks.add_parser('vsmart', help='detach template from vSmarts')
        vsmart_parser.set_defaults(template_filter=DeviceTemplateIndex.is_vsmart,
                                   set_title="vSmarts")

        # Parameters common to all sub-tasks
        for sub_task in (edge_parser, vsmart_parser):
            sub_task.add_argument('--templates', metavar='<regex>', type=regex_type,
                                  help='regular expression selecting templates to detach. Match on template name.')
            sub_task.add_argument('--devices', metavar='<regex>', type=regex_type,
                                  help='regular expression selecting devices to detach. Match on device name.')
            sub_task.add_argument('--reachable', action='store_true', help='select reachable devices only')
            sub_task.add_argument('--site', metavar='<id>', type=site_id_type, help='select devices with site ID')
            sub_task.add_argument('--system-ip', metavar='<ipv4>', type=ipv4_type, help='select device with system IP')
            sub_task.add_argument('--dryrun', action='store_true',
                                  help='dry-run mode. Attach operations are listed but nothing is pushed to vManage.')
            sub_task.add_argument('--batch', metavar='<size>', type=partial(int_type, 1, 9999),
                                  default=DEFAULT_BATCH_SIZE,
                                  help='maximum number of devices to include per vManage detach request '
                                       '(default: %(default)s)')

        return task_parser.parse_args(task_args)

    def runner(self, parsed_args, api: Optional[Rest] = None) -> Union[None, list]:
        self.is_dryrun = parsed_args.dryrun
        self.log_info(f'Detach templates task: vManage URL: "{api.base_url}"')

        try:
            matched_devices = dict(
                device_iter(api, parsed_args.devices, parsed_args.reachable, parsed_args.site, parsed_args.system_ip)
            )
            matched_templates = (
                (t_id, t_name)
                for t_id, t_name in DeviceTemplateIndex.get_raise(api).filtered_iter(parsed_args.template_filter)
                if parsed_args.templates is None or regex_search(parsed_args.templates, t_name)
            )
            reqs = self.detach(api, matched_templates, matched_devices, chunk_size=parsed_args.batch,
                               log_context=f"detaching {parsed_args.set_title}")
            if reqs:
                self.log_debug(f'Detach requests processed: {reqs}')
            else:
                self.log_info(f'No {parsed_args.set_title} detachments to process')

        except (RestAPIException, WaitActionsException) as ex:
            self.log_critical(f'Detach failed: {ex}')

        return


class DetachArgs(TaskArgs):
    templates: Optional[str] = None
    devices: Optional[str] = None
    site: Optional[str] = None
    system_ip: Optional[str] = None
    reachable: bool = False
    dryrun: bool = False
    batch: conint(ge=1, lt=9999) = DEFAULT_BATCH_SIZE

    # Validators
    _validate_regex = validator('templates', 'devices', allow_reuse=True)(validate_regex)
    _validate_site_id = validator('site', allow_reuse=True)(validate_site_id)
    _validate_ipv4 = validator('system_ip', allow_reuse=True)(validate_ipv4)


class DetachVsmartArgs(DetachArgs):
    template_filter: Callable = const(DeviceTemplateIndex.is_vsmart)
    set_title: str = const('vSmarts')


class DetachEdgeArgs(DetachArgs):
    template_filter: Callable = const(DeviceTemplateIndex.is_not_vsmart)
    set_title: str = const('WAN Edges')



"""
 Sastre - Automation Tools for Cisco SD-WAN Powered by Viptela

 cisco_sdwan.tasks.utils
 This module implements task utility classes and functions
"""
import os
import re
import argparse
from datetime import date
from getpass import getpass
from pathlib import Path
from cisco_sdwan.base.catalog import catalog_tags, op_catalog_tags, op_catalog_commands, CATALOG_TAG_ALL, OpType
from cisco_sdwan.base.models_base import filename_safe, DATA_DIR, ExtendedTemplate
from .common import Task

# Default local data store
DEFAULT_WORKDIR_FORMAT = 'backup_{address}_{date:%Y%m%d}'


def default_workdir(address):
    return DEFAULT_WORKDIR_FORMAT.format(date=date.today(), address=address or 'VMANAGE-ADDRESS')


class TaskOptions:
    _task_options = {}

    @classmethod
    def task(cls, task_str):
        task_cls = cls._task_options.get(task_str)
        if task_cls is None:
            raise argparse.ArgumentTypeError(f'Invalid task. Options are: {cls.options()}.')
        return task_cls

    @classmethod
    def options(cls):
        return ', '.join(cls._task_options)

    @classmethod
    def register(cls, task_name):
        """
        Decorator used for registering tasks.
        The class being decorated needs to be a subclass of Task.
        @param task_name: String presented to the user in order to select a task
        @return: decorator
        """

        def decorator(task_cls):
            if not isinstance(task_cls, type) or not issubclass(task_cls, Task):
                raise SastreException(f'Invalid task registration attempt: {task_cls.__name__}')

            cls._task_options[task_name] = task_cls
            return task_cls

        return decorator


class TagOptions:
    tag_options = catalog_tags() | {CATALOG_TAG_ALL}

    @classmethod
    def tag(cls, tag_str):
        if tag_str not in cls.tag_options:
            raise argparse.ArgumentTypeError(f'"{tag_str}" is not a valid tag. Available tags: {cls.options()}.')

        return tag_str

    @classmethod
    def tag_list(cls, tag_str_list):
        return [cls.tag(tag_str) for tag_str in tag_str_list]

    @classmethod
    def options(cls):
        return ', '.join(sorted(cls.tag_options, key=lambda x: '' if x == CATALOG_TAG_ALL else x))


class OpCmdOptions:
    @classmethod
    def tags(cls, op_type: OpType) -> str:
        return ', '.join(
            sorted(op_catalog_tags(op_type) | {CATALOG_TAG_ALL}, key=lambda x: '' if x == CATALOG_TAG_ALL else x)
        )

    @classmethod
    def commands(cls, op_type: OpType) -> str:
        return ', '.join(sorted(op_catalog_commands(op_type)))


class OpCmdSemantics(argparse.Action):
    # Using an action as opposed to a type check so that it can evaluate the full command line passed as opposed to
    # individual tokens.
    op_type: OpType = None

    def __call__(self, parser, namespace, values, option_string=None):
        full_command = ' '.join(values)
        pass_options = [
            len(values) == 1 and CATALOG_TAG_ALL in values,
            len(values) == 1 and set(values) <= op_catalog_tags(self.op_type),
            full_command in op_catalog_commands(self.op_type)
        ]
        if not any(pass_options):
            raise argparse.ArgumentError(self, f'"{full_command}" is not valid. '
                                               f'Group options: {OpCmdOptions.tags(self.op_type)}. '
                                               f'Command options: {OpCmdOptions.commands(self.op_type)}.')

        setattr(namespace, self.dest, values)


class RTCmdSemantics(OpCmdSemantics):
    op_type: OpType = OpType.RT


class StateCmdSemantics(OpCmdSemantics):
    op_type: OpType = OpType.STATE


class StatsCmdSemantics(OpCmdSemantics):
    op_type: OpType = OpType.STATS


def regex_type(regex_str):
    try:
        re.compile(regex_str)
    except (re.error, TypeError):
        if regex_str is not None:
            raise argparse.ArgumentTypeError(f'"{regex_str}" is not a valid regular expression.') from None

    return regex_str


def existing_workdir_type(workdir_str):
    if not Path(DATA_DIR, workdir_str).exists():
        raise argparse.ArgumentTypeError(f'Work directory "{workdir_str}" not found.')

    return workdir_str


def existing_file_type(filename_str):
    if not Path(filename_str).exists():
        raise argparse.ArgumentTypeError(f'File "{filename_str}" not found.')

    return filename_str


def filename_type(name_str):
    # Also allow . on filename, on top of what's allowed by filename_safe
    if re.sub(r'\.', '_', name_str) != filename_safe(name_str):
        raise argparse.ArgumentTypeError(
            f'Invalid name "{name_str}". Only alphanumeric characters, "-", "_", and "." are allowed.'
        )
    return name_str


def uuid_type(uuid_str):
    if re.match(r'[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$', uuid_str) is None:
        raise argparse.ArgumentTypeError(f'"{uuid_str}" is not a valid item ID.')

    return uuid_str


def non_empty_type(src_str):
    out_str = src_str.strip()
    if len(out_str) == 0:
        raise argparse.ArgumentTypeError('Value cannot be empty.')

    return out_str


def ipv4_type(ipv4_str):
    if re.match(r'\d+(?:\.\d+){3}$', ipv4_str) is None:
        raise argparse.ArgumentTypeError(f'"{ipv4_str}" is not a valid IPv4 address.')

    return ipv4_str


def site_id_type(site_id_str):
    try:
        site_id = int(site_id_str)
        if not 0 <= site_id <= 4294967295:
            raise ValueError()
    except ValueError:
        raise argparse.ArgumentTypeError(f'"{site_id_str}" is not a valid site-id.') from None

    return site_id_str


def version_type(version_str):
    # Development versions may follow this format: '20.1.999-98'
    if re.match(r'\d+([.-]\d+){1,3}$', version_str) is None:
        raise argparse.ArgumentTypeError(f'"{version_str}" is not a valid version identifier.')

    return '.'.join(([str(int(v)) for v in version_str.replace('-', '.').split('.')] + ['0', ])[:2])


def int_type(min_val, max_val, value_str):
    try:
        value_int = int(value_str)
        if not min_val <= value_int <= max_val:
            raise ValueError()
    except ValueError:
        raise argparse.ArgumentTypeError(f'Invalid value: "{value_str}". Must be an integer between '
                                         f'{min_val} and {max_val}, inclusive.') from None

    return value_int


class EnvVar(argparse.Action):
    def __init__(self, nargs=None, envvar=None, required=True, default=None, **kwargs):
        if nargs is not None:
            raise ValueError('nargs not allowed')
        if envvar is None:
            raise ValueError('envvar is required')

        default = os.environ.get(envvar) or default
        required = required and default is None
        super().__init__(default=default, required=required, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values)


class PromptArg:
    def __init__(self, argument, prompt, secure_prompt=False, validate=non_empty_type):
        self.argument = argument
        self.prompt = prompt
        self.prompt_func = getpass if secure_prompt else input
        self.validate = validate

    def __call__(self):
        while True:
            try:
                value = self.validate(self.prompt_func(self.prompt))
            except argparse.ArgumentTypeError as ex:
                print(f'{ex} Please try again, or ^C to terminate.')
            else:
                return value


def ext_template_type(template_str):
    try:
        ExtendedTemplate(template_str)('test')
    except ValueError as ex:
        raise argparse.ArgumentTypeError(ex) from None

    return template_str


class SastreException(Exception):
    """ Exception for main app errors """
    pass
